import abc
import copy
import functools
from typing import Optional, Any, TypeVar, Generic, TYPE_CHECKING

from PySide6.QtCore import (
    QAbstractTableModel,
    QObject,
    QModelIndex,
    QAbstractListModel,
    Qt,
    QSize,
    QSettings,
)
from PySide6.QtGui import QAction, QBrush, QColor, QFont
from PySide6.QtWidgets import QMenu, QColorDialog

import caqtus.gui.qtutil.qabc as qabc
from caqtus.session.shot import TimeLane
from caqtus.types.timelane import TimeLanes
from caqtus.types.expression import Expression

if TYPE_CHECKING:
    from .extension import CondetrolLaneExtensionProtocol


class TimeStepNameModel(QAbstractListModel):
    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)
        self._names: list[str] = []

    def set_names(self, names: list[str]):
        self.beginResetModel()
        self._names = copy.deepcopy(names)
        self.endResetModel()

    def get_names(self) -> list[str]:
        return copy.deepcopy(self._names)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._names)

    def data(
        self, index: QModelIndex, role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole
    ):
        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
            return self._names[index.row()]
        if role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            if not isinstance(value, str):
                raise TypeError(f"Expected str, got {type(value)}")
            self._names[index.row()] = value
            self.dataChanged.emit(index, index)
            return True
        return False

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsEditable

    def insertRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        if not (0 <= row <= self.rowCount()):
            return False
        self.beginInsertRows(parent, row, row)
        self._names.insert(row, f"Step {row}")
        self.endInsertRows()
        return True

    def removeRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        if not (0 <= row < self.rowCount()):
            return False
        self.beginRemoveRows(parent, row, row)
        del self._names[row]
        self.endRemoveRows()
        return True

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return "Step names"
            elif orientation == Qt.Orientation.Vertical:
                return section
        elif role == Qt.ItemDataRole.FontRole:
            font = QFont()
            font.setBold(True)
            return font


class TimeStepDurationModel(QAbstractListModel):
    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)
        self._durations: list[Expression] = []

    def set_durations(self, durations: list[Expression]):
        self.beginResetModel()
        self._durations = copy.deepcopy(durations)
        self.endResetModel()

    def get_duration(self) -> list[Expression]:
        return copy.deepcopy(self._durations)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._durations)

    def data(
        self, index: QModelIndex, role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole
    ):
        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
            return self._durations[index.row()].body
        if role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            if not isinstance(value, str):
                raise TypeError(f"Expected str, got {type(value)}")
            self._durations[index.row()].body = value
            self.dataChanged.emit(index, index)
            return True
        return False

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsEditable

    def insertRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        if not (0 <= row <= self.rowCount()):
            return False
        self.beginInsertRows(parent, row, row)
        self._durations.insert(row, Expression("..."))
        self.endInsertRows()
        return True

    def removeRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        if not (0 <= row < self.rowCount()):
            return False
        self.beginRemoveRows(parent, row, row)
        del self._durations[row]
        self.endRemoveRows()
        return True

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return "Step durations"
            elif orientation == Qt.Orientation.Vertical:
                return section
        elif role == Qt.ItemDataRole.FontRole:
            font = QFont()
            font.setBold(True)
            return font


T = TypeVar("T")
L = TypeVar("L", bound=TimeLane)
O = TypeVar("O", bound=Any)


class TimeLaneModel(QAbstractListModel, Generic[L, O], metaclass=qabc.QABCMeta):
    """An abstract list model to represent a time lane.

    This class inherits from :class:`PySide6.QtCore.QAbstractListModel` and can be
    used to represent a lane in the timelanes editor.

    It is meant to be subclassed for each lane type that needs to be represented in
    the timelanes editor.
    Some common methods are implemented here, but subclasses will need to implement at
    least the abstract methods: :meth:`data`, :meth:`setData`, :meth:`insertRow`.
    In addition, subclasses may want to override :meth:`flags` to change the item flags
    for the cells in the lane.
    The :meth:`get_cell_context_actions` method can be overridden to add context menu
    actions to the cells in the lane.
    """

    def __init__(self, name: str, lane: L, parent: Optional[QObject] = None):
        super().__init__(parent)
        self._name = name
        self._lane = lane

    def get_lane(self) -> L:
        """Return a copy of the lane represented by this model."""

        return copy.deepcopy(self._lane)

    def set_lane(self, lane: L) -> None:
        """Set the lane represented by this model."""

        self.beginResetModel()
        self._lane = copy.deepcopy(lane)
        self.endResetModel()

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        """Return the number of steps in the lane."""

        return len(self._lane)

    @abc.abstractmethod
    def data(
        self, index: QModelIndex, role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole
    ) -> Any:
        """Return the data for the given index and role.

        See :meth:`PySide6.QtCore.QAbstractItemModel.data` for more information.
        """

        return None

    @abc.abstractmethod
    def setData(self, index, value, role=Qt.ItemDataRole.EditRole) -> bool:
        """Set the data for the given index and role.

        See :meth:`PySide6.QtCore.QAbstractItemModel.setData` for more information.
        """

        raise NotImplementedError

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole,
    ):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return self._name
            elif orientation == Qt.Orientation.Vertical:
                return section
        elif role == Qt.ItemDataRole.FontRole:
            font = QFont()
            font.setBold(True)
            return font

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        """Return the flags for the given index.

        By default, the flags are set to `ItemIsEnabled`, `ItemIsEditable`, and
        `ItemIsSelectable`.
        """

        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        return (
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsSelectable
        )

    @abc.abstractmethod
    def insertRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        raise NotImplementedError

    def insert_value(self, row: int, value: T) -> bool:
        if not (0 <= row <= len(self._lane)):
            return False
        self.beginInsertRows(QModelIndex(), row, row)
        if row == len(self._lane):
            self._lane.append(value)
        else:
            start, stop = self._lane.get_bounds(row)
            self._lane.insert(row, value)
            if start < row < stop:
                self._lane[start : stop + 1] = self._lane[start]
        self.endInsertRows()
        return True

    def removeRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        if not (0 <= row < len(self._lane)):
            return False
        self.beginRemoveRows(parent, row, row)
        del self._lane[row]
        self.endRemoveRows()
        return True

    def get_cell_context_actions(self, index: QModelIndex) -> list[QAction | QMenu]:
        break_span_action = QAction("Break block")
        break_span_action.triggered.connect(lambda: self.break_span(index))
        return [break_span_action]

    def span(self, index) -> QSize:
        start, stop = self._lane.get_bounds(index.row())
        if index.row() == start:
            return QSize(1, stop - start)
        else:
            return QSize(1, 1)

    def break_span(self, index: QModelIndex) -> bool:
        start, stop = self._lane.get_bounds(index.row())
        value = self._lane[index.row()]
        for i in range(start, stop):
            self._lane[i] = value
        self.dataChanged.emit(self.index(start), self.index(stop - 1))
        return True

    def expand_step(self, step: int, start: int, stop: int) -> None:
        value = self._lane[step]
        self._lane[start : stop + 1] = value
        self.dataChanged.emit(self.index(start), self.index(stop - 1))

    def get_header_context_actions(self) -> list[QAction | QMenu]:
        """Return a list of context menu actions for the lane header."""

        return []

    def simplify(self) -> None:
        """Simplify the lane by merging contiguous blocks of the same value."""

        self.beginResetModel()
        start = 0
        for i in range(1, len(self._lane)):
            if self._lane[i] != self._lane[start]:
                self._lane[start:i] = self._lane[start]
                start = i
        self._lane[start:] = self._lane[start]
        self.endResetModel()


class ColoredTimeLaneModel(TimeLaneModel[L, O], metaclass=qabc.QABCMeta):
    """A time lane model that can be colored.

    Instances of this class can be used to color the cells in a lane.
    They have the attribute :attr:`_brush` that is optionally a :class:`QBrush` that
    can be used to color the cells in the lane.
    """

    def __init__(self, name: str, lane: TimeLane, parent: Optional[QObject] = None):
        super().__init__(name, lane, parent)
        self._brush: Optional[QBrush] = None

        color = QSettings().value(f"lane color/{self._name}", None)
        if color is not None:
            self._brush = QBrush(color)
        else:
            self._brush = None

    @abc.abstractmethod
    def data(
        self, index: QModelIndex, role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole
    ) -> Any:
        """Returns its brush for the `Qt.ItemDataRole.ForegroundRole` role."""

        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.ForegroundRole:
            return self._brush
        return super().data(index, role)

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole,
    ):
        if role == Qt.ItemDataRole.ForegroundRole:
            return self._brush
        return super().headerData(section, orientation, role)

    def get_header_context_actions(self) -> list[QAction | QMenu]:
        action = QAction("Change color")
        action.triggered.connect(lambda: self._change_color())
        return [action]

    def _change_color(self):
        if self._brush is None:
            color = QColorDialog.getColor(title=f"Select color for {self._name}")
        else:
            color = QColorDialog.getColor(
                self._brush.color(), title=f"Select color for {self._name}"
            )
        if color.isValid():
            self.setHeaderData(
                0, Qt.Orientation.Horizontal, color, Qt.ItemDataRole.ForegroundRole
            )

    def setHeaderData(self, section, orientation, value, role=Qt.ItemDataRole.EditRole):
        change = False
        if (
            orientation == Qt.Orientation.Horizontal
            and role == Qt.ItemDataRole.ForegroundRole
        ):
            if isinstance(value, QColor):
                self._brush = QBrush(value)
                settings = QSettings()
                settings.setValue(f"lane color/{self._name}", value)
                change = True
            elif value is None:
                self._brush = None
                settings = QSettings()
                settings.remove(f"lane color/{self._name}")
                change = True
        if change:
            self.headerDataChanged.emit(orientation, section, section)
            return True

        return super().setHeaderData(section, orientation, value, role)


class TimeLanesModel(QAbstractTableModel, metaclass=qabc.QABCMeta):
    def __init__(
        self,
        extension: "CondetrolLaneExtensionProtocol",
        parent: Optional[QObject] = None,
    ):
        super().__init__(parent)
        self._step_names_model = TimeStepNameModel(self)
        self._step_durations_model = TimeStepDurationModel(self)
        self._lane_models: list[TimeLaneModel] = []
        self._extension = extension

        self._step_names_model.dataChanged.connect(self.on_step_names_data_changed)
        self._step_durations_model.dataChanged.connect(
            self.on_step_durations_data_changed
        )
        self._read_only = False

    def set_read_only(self, read_only: bool) -> None:
        self._read_only = read_only

    def is_read_only(self) -> bool:
        return self._read_only

    def on_step_names_data_changed(
        self,
        top_left: QModelIndex,
        bottom_right: QModelIndex,
        roles: list[Qt.ItemDataRole],
    ):
        self.dataChanged.emit(
            self.index(0, top_left.row()), self.index(0, bottom_right.row())
        )

    def on_step_durations_data_changed(
        self,
        top_left: QModelIndex,
        bottom_right: QModelIndex,
        roles: list[Qt.ItemDataRole],
    ):
        self.dataChanged.emit(
            self.index(1, top_left.row()), self.index(1, bottom_right.row())
        )

    def set_timelanes(self, timelanes: TimeLanes):
        # Don't check if read only, because we need to update the content of the editor
        # even if it is readonly when swapping sequences.
        new_models = []
        for index, (name, lane) in enumerate(timelanes.lanes.items()):
            lane_model = self.create_lane_model(name, lane)
            new_models.append(lane_model)

        self.beginResetModel()
        self._step_names_model.set_names(timelanes.step_names)
        self._step_durations_model.set_durations(timelanes.step_durations)
        self._lane_models.clear()
        self._lane_models.extend(new_models)
        self.endResetModel()

    def create_lane_model(self, name: str, lane: TimeLane) -> TimeLaneModel:
        lane_model = self._extension.get_lane_model(lane, name)
        lane_model.setParent(self)
        lane_model.set_lane(lane)
        lane_model.dataChanged.connect(
            # For some reason, functools.partial does not work here, but lambda does.
            # functools.partial(
            #     self.on_lane_model_data_changed, lane_model=lane_model
            # )
            lambda top_left, bottom_right: self.on_lane_model_data_changed(
                top_left, bottom_right, lane_model
            )
        )
        lane_model.headerDataChanged.connect(
            functools.partial(self.on_lane_header_data_changed, lane_model=lane_model)
        )
        return lane_model

    def on_lane_model_data_changed(
        self,
        top_left: QModelIndex,
        bottom_right: QModelIndex,
        lane_model: TimeLaneModel,
    ):
        lane_index = self._lane_models.index(lane_model)
        self.dataChanged.emit(
            self.index(lane_index + 2, top_left.row()),
            self.index(lane_index + 2, bottom_right.row()),
        )

    def on_lane_header_data_changed(
        self,
        orientation: Qt.Orientation,
        first: int,
        last: int,
        lane_model: TimeLaneModel,
    ):
        lane_index = self._lane_models.index(lane_model)
        if orientation == Qt.Orientation.Horizontal:
            self.headerDataChanged.emit(
                Qt.Orientation.Vertical,
                lane_index + 2,
                lane_index + 2,
            )

    def insert_time_lane(
        self, name: str, timelane: TimeLane, index: Optional[int] = None
    ):
        if self._read_only:
            return
        if index is None:
            index = len(self._lane_models)
        if not (0 <= index <= len(self._lane_models)):
            raise IndexError(f"Index {index} is out of range")
        if len(timelane) != self.columnCount():
            raise ValueError(
                f"Length of time lane ({len(timelane)}) does not match "
                f"number of columns ({self.columnCount()})"
            )
        already_used_names = {
            model.headerData(0, Qt.Orientation.Horizontal)
            for model in self._lane_models
        }
        if name in already_used_names:
            raise ValueError(f"Name {name} is already used")
        lane_model = self.create_lane_model(name, timelane)
        self.beginInsertRows(QModelIndex(), index, index)
        self._lane_models.insert(index, lane_model)
        self.endInsertRows()

    def get_lane(self, index: int) -> TimeLane:
        return self._lane_models[index].get_lane()

    def get_lane_name(self, index: int) -> str:
        return self._lane_models[index].headerData(0, Qt.Orientation.Horizontal)

    def get_timelanes(self) -> TimeLanes:
        return TimeLanes(
            step_names=self._step_names_model.get_names(),
            step_durations=self._step_durations_model.get_duration(),
            lanes={
                model.headerData(0, Qt.Orientation.Horizontal): model.get_lane()
                for model in self._lane_models
            },
        )

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        count = self._step_names_model.rowCount()
        assert count == self._step_durations_model.rowCount()
        assert all(model.rowCount() == count for model in self._lane_models), [
            model.rowCount() for model in self._lane_models
        ]
        return count

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._lane_models) + 2

    def data(self, index, role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        return self._map_to_source(index).data(role)

    def setData(self, index, value, role: Qt.ItemDataRole = Qt.ItemDataRole.EditRole):
        if self._read_only:
            return False
        if not index.isValid():
            return False
        mapped_index = self._map_to_source(index)
        return mapped_index.model().setData(mapped_index, value, role)

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        mapped_index = self._map_to_source(index)
        flags = mapped_index.model().flags(mapped_index)
        if self._read_only:
            flags &= ~Qt.ItemFlag.ItemIsEditable
            flags &= ~Qt.ItemFlag.ItemIsDropEnabled
        return flags

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole,
    ):
        if orientation == Qt.Orientation.Horizontal:
            if role == Qt.ItemDataRole.DisplayRole:
                return f"Step {section}"
        elif orientation == Qt.Orientation.Vertical:
            if section == 0:
                return self._step_names_model.headerData(
                    0, Qt.Orientation.Horizontal, role
                )
            elif section == 1:
                return self._step_durations_model.headerData(
                    0, Qt.Orientation.Horizontal, role
                )
            else:
                return self._lane_models[section - 2].headerData(
                    0, Qt.Orientation.Horizontal, role
                )

    def insertColumn(self, column, parent: QModelIndex = QModelIndex()) -> bool:
        if self._read_only:
            return False
        if not (0 <= column <= self.columnCount()):
            return False
        self.beginInsertColumns(parent, column, column)
        self._step_names_model.insertRow(column)
        self._step_durations_model.insertRow(column)
        for lane_model in self._lane_models:
            lane_model.insertRow(column)
        self.endInsertColumns()
        self.modelReset.emit()

        return True

    def removeColumn(self, column, parent: QModelIndex = QModelIndex()) -> bool:
        if self._read_only:
            return False
        if not (0 <= column < self.columnCount()):
            return False
        self.beginRemoveColumns(parent, column, column)
        self._step_names_model.removeRow(column)
        self._step_durations_model.removeRow(column)
        for lane_model in self._lane_models:
            lane_model.removeRow(column)
        self.endRemoveColumns()
        self.modelReset.emit()
        return True

    def removeRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        if self._read_only:
            return False
        if not (2 <= row < self.rowCount()):
            return False
        self.beginRemoveRows(parent, row, row)
        del self._lane_models[row - 2]
        self.endRemoveRows()

    def get_cell_context_actions(self, index: QModelIndex) -> list[QAction | QMenu]:
        if not index.isValid():
            return []
        if self._read_only:
            return []
        if index.row() >= 2:
            return self._lane_models[index.row() - 2].get_cell_context_actions(
                self._map_to_source(index)
            )
        else:
            return []

    def get_lane_header_context_actions(self, lane_index: int) -> list[QAction | QMenu]:
        if not 0 <= lane_index < len(self._lane_models):
            return []
        if self._read_only:
            return []
        return self._lane_models[lane_index].get_header_context_actions()

    def span(self, index):
        if not index.isValid():
            return QSize(1, 1)
        if index.row() >= 2:
            mapped_index = self._map_to_source(index)
            span = self._lane_models[index.row() - 2].span(mapped_index)
            return QSize(span.height(), span.width())
        return QSize(1, 1)

    def expand_step(self, step: int, lane_index: int, start: int, stop: int):
        if self._read_only:
            return
        lane_model = self._lane_models[lane_index]
        lane_model.expand_step(step, start, stop)

    def _map_to_source(self, index: QModelIndex) -> QModelIndex:
        assert index.isValid()
        assert self.hasIndex(index.row(), index.column())
        if index.row() == 0:
            return self._step_names_model.index(index.column(), 0)
        elif index.row() == 1:
            return self._step_durations_model.index(index.column(), 0)
        else:
            return self._lane_models[index.row() - 2].index(index.column(), 0)

    def simplify(self) -> None:
        self.beginResetModel()
        for lane_model in self._lane_models:
            lane_model.simplify()
        self.endResetModel()



from collections.abc import Iterable
from typing import TypeVar

T = TypeVar("T")


def first(iterable: Iterable[T]) -> T:
    """Returns the first value of an iterable.

    Raises:
        ValueError if the iterable is empty.
    """

    iterator = iter(iterable)
    try:
        return next(iterator)
    except StopIteration:
        raise ValueError("Iterable is empty")



from __future__ import annotations

from enum import Enum

from ..data_type import DataType


class State(Enum):
    """Indicate the state of a sequence.

    Each sequence in the experiment session has a state. Only sequences in the DRAFT
    state can have their configuration modified. Only sequences in the RUNNING state
    can have data added to them.

    Attributes:
        DRAFT: The sequence was not started yet and is still being edited. When the
            sequence is passed to the experiment manager, it will transition to the
            PREPARING state.
        PREPARING: The sequence is being prepared to run and is currently acquiring the
            necessary resources and devices. No shot have been scheduled to run yet. The
            sequence cannot be interrupted while it is preparing. If the preparation
            succeeds, the sequence will transition to the RUNNING state. If the
            preparation fails, the sequence will transition to the CRASHED state.
        RUNNING: The sequence is currently running shots. When all shots are completed,
            the sequence will transition to the FINISHED state. If the sequence is
            interrupted, it will transition to the INTERRUPTED state. If an error occurs
            while running the sequence, it will transition to the CRASHED state.
        FINISHED: The sequence was successfully run and all shots were completed.
        INTERRUPTED: The sequence was interrupted by the user before all shots were
        completed.
        CRASHED: An error occurred while running the sequence.

    The states FINISHED, INTERRUPTED and CRASHED are terminal states. Once a sequence is
    in one of these states, the only way to change its state is to erase all associated
    data and reset it back to the DRAFT state.
    """

    DRAFT = "draft"
    PREPARING = "preparing"
    RUNNING = "running"
    FINISHED = "finished"
    INTERRUPTED = "interrupted"
    CRASHED = "crashed"

    @classmethod
    def is_transition_allowed(cls, old_state: State, new_state: State) -> bool:
        return new_state in _ALLOWED_TRANSITIONS[old_state]

    def is_editable(self) -> bool:
        """Indicate if a sequence in this state can be edited."""

        return self in {State.DRAFT}

    def can_add_data(self, data_type: "DataType") -> bool:
        """Indicate if a sequence in this state can add data of the given type."""

        return self in {State.RUNNING}

    def __str__(self):
        return self.value


_ALLOWED_TRANSITIONS = {
    State.DRAFT: {State.PREPARING},
    State.PREPARING: {State.RUNNING, State.CRASHED},
    State.RUNNING: {State.FINISHED, State.INTERRUPTED, State.CRASHED},
    State.FINISHED: {State.DRAFT},
    State.INTERRUPTED: {State.DRAFT},
    State.CRASHED: {State.DRAFT},
}



from ._extension import TimeLaneExtension
from ...gui.condetrol.timelanes_editor import TimeLaneModel, DigitalTimeLaneModel
from ...gui.condetrol.timelanes_editor.digital_lane_delegate import (
    DigitalTimeLaneDelegate,
)
from ...types.timelane import DigitalTimeLane
from ...utils import serialization
from ...utils.serialization import JSON


def create_digital_lane(number_steps: int) -> DigitalTimeLane:
    return DigitalTimeLane([False] * number_steps)


def create_lane_model(lane, name: str) -> TimeLaneModel:
    model = DigitalTimeLaneModel(name)
    model.set_lane(lane)
    return model


def create_lane_delegate(lane, name: str) -> DigitalTimeLaneDelegate:
    return DigitalTimeLaneDelegate()


def dump_digital_lane(time_lane: DigitalTimeLane):
    return serialization.converters["json"].unstructure(time_lane, DigitalTimeLane)


def load_digital_lane(content: JSON):
    return serialization.converters["json"].structure(content, DigitalTimeLane)


digital_time_lane_extension = TimeLaneExtension(
    label="Digital",
    lane_type=DigitalTimeLane,
    dumper=dump_digital_lane,
    loader=load_digital_lane,
    lane_factory=create_digital_lane,
    lane_model_factory=create_lane_model,
    lane_delegate_factory=create_lane_delegate,
    type_tag="digital",
)



__version__ = "0.1.0"

from . import base
from .unit_namespace import units
from .units import (
    ureg,
    unit_registry,
    Quantity,
    Unit,
    UndefinedUnitError,
    DimensionalityError,
    dimensionless,
    TIME_UNITS,
    FREQUENCY_UNITS,
    POWER_UNITS,
    DIMENSIONLESS_UNITS,
    CURRENT_UNITS,
    VOLTAGE_UNITS,
    UNITS,
    UnitLike,
    InvalidDimensionalityError,
)

__all__ = [
    "__version__",
    "ureg",
    "unit_registry",
    "Quantity",
    "Unit",
    "UndefinedUnitError",
    "DimensionalityError",
    "dimensionless",
    "TIME_UNITS",
    "FREQUENCY_UNITS",
    "POWER_UNITS",
    "DIMENSIONLESS_UNITS",
    "CURRENT_UNITS",
    "VOLTAGE_UNITS",
    "UNITS",
    "units",
    "UnitLike",
    "InvalidDimensionalityError",
    "base",
]



import functools
from collections.abc import Callable
from typing import Optional, TypeAlias, Protocol, TypeVar, Any

from PySide6.QtWidgets import QStyledItemDelegate

from caqtus.session.shot import (
    TimeLane,
)
from caqtus.types.timelane import TimeLanes
from caqtus.types.timelane.serializer import TimeLaneSerializer
from caqtus.utils import serialization
from ._protocol import CondetrolLaneExtensionProtocol
from ..model import TimeLaneModel

L = TypeVar("L", bound=TimeLane)

LaneFactory: TypeAlias = Callable[[int], L]


class LaneDelegateFactory(Protocol[L]):
    """A factory for lane delegates."""

    def __call__(
        self,
        lane: L,
        lane_name: str,
    ) -> Optional[QStyledItemDelegate]:
        """Create a delegate for the lane passed as argument."""
        ...


class LaneModelFactory(Protocol[L]):
    def __call__(
        self,
        lane: L,
        lane_name: str,
    ) -> TimeLaneModel[L, Any]:
        """Create a delegate for the lane passed as argument."""
        ...


class CondetrolLaneExtension(CondetrolLaneExtensionProtocol):
    def __init__(self):
        self.get_lane_delegate = functools.singledispatch(default_lane_delegate_factory)
        self.get_lane_model = functools.singledispatch(default_lane_model_factory)
        self._lane_factories: dict[str, LaneFactory] = {}
        self._lane_serializer = TimeLaneSerializer()

    def set_lane_serializer(self, serializer: TimeLaneSerializer) -> None:
        self._lane_serializer = serializer

    def register_lane_factory(self, lane_label: str, factory: LaneFactory) -> None:
        self._lane_factories[lane_label] = factory

    def register_lane_delegate_factory(
        self, lane_type: type[L], factory: LaneDelegateFactory[L]
    ) -> None:
        self.get_lane_delegate.register(lane_type)(factory)

    def register_lane_model_factory(
        self, lane_type: type[L], factory: LaneModelFactory[L]
    ) -> None:
        self.get_lane_model.register(lane_type)(factory)

    def available_new_lanes(self) -> set[str]:
        return set(self._lane_factories.keys())

    def create_new_lane(self, lane_label: str, steps: int) -> TimeLane:
        lane = self._lane_factories[lane_label](steps)
        if not isinstance(lane, TimeLane):
            raise TypeError(f"Expected a TimeLane, got {type(lane)}.")
        if len(lane) != steps:
            raise ValueError(
                f"Expected a lane with {steps} steps, got {len(lane)} steps."
            )
        return lane

    def unstructure_time_lanes(self, time_lanes: TimeLanes) -> serialization.JSON:
        return self._lane_serializer.unstructure_time_lanes(time_lanes)

    def structure_time_lanes(self, content: serialization.JSON) -> TimeLanes:
        return self._lane_serializer.structure_time_lanes(content)


def default_lane_model_factory(lane, name: str) -> TimeLaneModel:
    if not isinstance(lane, TimeLane):
        raise TypeError(f"Expected a TimeLane, got {type(lane)}.")

    raise NotImplementedError(f"Don't know how to provide a model for {type(lane)}.")


def default_lane_delegate_factory(
    lane,
    lane_name: str,
) -> Optional[QStyledItemDelegate]:
    return None



import functools
from typing import Optional

from PySide6.QtWidgets import QWidget, QVBoxLayout

from caqtus.device.sequencer.channel_commands import (
    ChannelOutput,
    Constant,
    DeviceTrigger,
    LaneValues,
    CalibratedAnalogMapping,
)
from caqtus.device.sequencer.channel_commands.timing import Advance, BroadenLeft
from caqtus.gui.common.NodeGraphQt import NodeGraph, BaseNode, NodesPaletteWidget
from ._analog_mapping_node import CalibratedAnalogMappingNode
from ._constant_node import ConstantNode
from ._device_trigger_node import DeviceTriggerNode
from ._lane_node import LaneNode
from ._output_node import OutputNode
from ._timing_nodes import AdvanceNode, BroadenLeftNode


class ChannelOutputEditor(QWidget):
    def __init__(self, channel_output: ChannelOutput, parent: Optional[QWidget] = None):
        super().__init__(parent)

        self.graph = NodeGraph(self)
        self.graph.register_node(ConstantNode)
        self.graph.register_node(DeviceTriggerNode)
        self.graph.register_node(LaneNode)
        self.graph.register_node(AdvanceNode)
        self.graph.register_node(CalibratedAnalogMappingNode)
        self.graph.register_node(BroadenLeftNode)
        self.nodes_tree = NodesPaletteWidget(node_graph=self.graph, parent=self)
        self.nodes_tree.set_category_label("caqtus.sequencer_node.source", "Source")
        self.nodes_tree.set_category_label("caqtus.sequencer_node.timing", "Timing")
        self.nodes_tree.set_category_label("caqtus.sequencer_node.mapping", "Mapping")

        layout = QVBoxLayout(self)
        layout.addWidget(self.graph.widget, 1)
        layout.addWidget(self.nodes_tree, 0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.output_node = OutputNode()
        self.graph.add_node(
            self.output_node, selected=False, pos=[0, 0], push_undo=False
        )

        node = self.build_node(channel_output)
        node.outputs()["out"].connect_to(self.output_node.inputs()["in"])
        self.graph.auto_layout_nodes()
        self.graph.set_pipe_collision(True)
        self.graph.clear_undo_stack()

    def get_output(self) -> ChannelOutput:
        connected_node = self.output_node.connected_node()
        if connected_node is None:
            raise MissingInputError("No node is connected to the output node")
        output = construct_output(connected_node)
        return output

    @functools.singledispatchmethod
    def build_node(self, channel_output: ChannelOutput) -> BaseNode:
        raise NotImplementedError(f"Can't build node from {type(channel_output)}")

    @build_node.register
    def build_constant(self, constant: Constant) -> ConstantNode:
        node = ConstantNode()
        node.set_value(constant.value)
        self.graph.add_node(node, selected=False, push_undo=False)
        return node

    @build_node.register
    def build_device_trigger_node(
        self, device_trigger: DeviceTrigger
    ) -> DeviceTriggerNode:
        node = DeviceTriggerNode()
        node.set_device_name(device_trigger.device_name)
        self.graph.add_node(node, selected=False, push_undo=False)
        if device_trigger.default is not None:
            default_node = self.build_node(device_trigger.default)
            default_node.outputs()["out"].connect_to(node.default_port)
        return node

    @build_node.register
    def build_lane_node(self, lane_values: LaneValues) -> LaneNode:
        node = LaneNode()
        node.set_lane_name(lane_values.lane)
        self.graph.add_node(node, selected=False, push_undo=False)
        if lane_values.default is not None:
            default_node = self.build_node(lane_values.default)
            default_node.outputs()["out"].connect_to(node.default_port)
        return node

    @build_node.register
    def build_advance_node(self, advance: Advance) -> AdvanceNode:
        node = AdvanceNode()
        node.set_advance(advance.advance)
        self.graph.add_node(node, selected=False, push_undo=False)
        input_node = self.build_node(advance.input_)
        input_node.outputs()["out"].connect_to(node.input_port)
        return node

    @build_node.register
    def build_analog_mapping_node(
        self, analog_mapping: CalibratedAnalogMapping
    ) -> CalibratedAnalogMappingNode:
        node = CalibratedAnalogMappingNode()
        node.set_units(analog_mapping.input_units, analog_mapping.output_units)
        node.set_data_points(analog_mapping.measured_data_points)
        self.graph.add_node(node, selected=False, push_undo=False)
        input_node = self.build_node(analog_mapping.input_)
        input_node.outputs()["out"].connect_to(node.input_port)
        return node

    @build_node.register
    def build_broaden_left_node(self, broaden_left: BroadenLeft) -> BroadenLeftNode:
        node = BroadenLeftNode()
        node.set_width(broaden_left.width)
        self.graph.add_node(node, selected=False, push_undo=False)
        input_node = self.build_node(broaden_left.input_)
        input_node.outputs()["out"].connect_to(node.input_port)
        return node


@functools.singledispatch
def construct_output(node) -> ChannelOutput:
    raise NotImplementedError(f"Cant construct output from {type(node)}")


@construct_output.register
def construct_constant(node: ConstantNode) -> Constant:
    return Constant(value=node.get_value())


@construct_output.register
def construct_device_trigger(node: DeviceTriggerNode) -> DeviceTrigger:
    device_name = node.get_device_name()
    default_node = node.get_default_node()
    if default_node is None:
        default = None
    else:
        default = construct_output(default_node)
    return DeviceTrigger(device_name=device_name, default=default)


@construct_output.register
def construct_lane_values(node: LaneNode) -> LaneValues:
    lane_name = node.get_lane_name()
    default_node = node.get_default_node()
    if default_node is None:
        default = None
    else:
        default = construct_output(default_node)
    return LaneValues(lane=lane_name, default=default)


@construct_output.register
def construct_advance(node: AdvanceNode) -> Advance:
    advance = node.get_advance()
    input_node = node.get_input_node()
    if input_node is None:
        raise MissingInputError(f"Advance node {node.name()} must have an input node")
    else:
        input_ = construct_output(input_node)
    return Advance(advance=advance, input_=input_)


@construct_output.register
def construct_analog_mapping(
    node: CalibratedAnalogMappingNode,
) -> CalibratedAnalogMapping:
    input_node = node.get_input_node()
    if input_node is None:
        raise MissingInputError(
            f"Analog mapping node {node.name()} must have an input node"
        )
    else:
        input_ = construct_output(input_node)
    input_units, output_units = node.get_units()
    return CalibratedAnalogMapping(
        input_=input_,
        input_units=input_units,
        output_units=output_units,
        measured_data_points=tuple(node.get_data_points()),
    )


@construct_output.register
def construct_broaden_left(node: BroadenLeftNode) -> BroadenLeft:
    width = node.get_width()
    input_node = node.get_input_node()
    if input_node is None:
        raise MissingInputError(
            f"Broaden left node {node.name()} must have an input node"
        )
    else:
        input_ = construct_output(input_node)
    return BroadenLeft(width=width, input_=input_)


class InvalidNodeConfigurationError(ValueError):
    pass


class MissingInputError(InvalidNodeConfigurationError):
    pass



import abc
from collections.abc import MutableMapping

from caqtus.device import DeviceName, DeviceConfiguration


class DeviceConfigurationCollection(
    MutableMapping[DeviceName, DeviceConfiguration], abc.ABC
):
    pass



"""This module provides widgets to edit a sequencer configuration."""

from .channels_widget import SequencerChannelWidget
from .sequencer_configuration_editor import SequencerConfigurationEditor, TimeStepEditor

__all__ = [
    "SequencerConfigurationEditor",
    "TimeStepEditor",
    "SequencerChannelWidget",
]



"""This module contains classes and functions to manage devices."""

from ._controller import DeviceController
from .configuration import (
    DeviceConfiguration,
    get_configurations_by_type,
    DeviceParameter,
)
from ._name import DeviceName
from .runtime import Device, RuntimeDevice

__all__ = [
    "DeviceName",
    "DeviceConfiguration",
    "DeviceParameter",
    "Device",
    "RuntimeDevice",
    "get_configurations_by_type",
    "DeviceController",
]



import logging

import anyio
import anyio.to_thread
import polars
import qtawesome
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication, QMainWindow, QDockWidget
from caqtus.analysis.loading import DataImporter
from caqtus.gui.common.sequence_hierarchy import AsyncPathHierarchyView
from caqtus.session import ExperimentSessionMaker

from .data_loading import DataLoader
from .graphplot_main_window_ui import Ui_GraphPlotMainWindow
from .views.error_bar_view import ErrorBarView
from ..qtutil import qt_trio

logger = logging.getLogger(__name__)


class GraphPlot:
    def __init__(
        self, data_importer: DataImporter, session_maker: ExperimentSessionMaker, *args
    ) -> None:
        """
        Args:
            data_importer: A callable used to import data from shots.
            session_maker: A callable used to create sessions from which the application can retrieve data.
        """

        self.app = QApplication(*args)
        self.app.setApplicationName("GraphPlot")
        self.app.setStyle("Fusion")
        self.app.setWindowIcon(qtawesome.icon("mdi6.chart-line", size=64))
        self.main_window = GraphPlotMainWindow(data_importer, session_maker)

    def run(self) -> None:
        self.main_window.show()
        qt_trio.run(self.main_window.start)


class GraphPlotMainWindow(QMainWindow, Ui_GraphPlotMainWindow):
    """The main window for the GraphPlot application.

    On the left, it displays a tree view of the experiment session's sequences.
    On the right, there is a widget to define how to import data from the sequences.
    In the middle, there is a view of the data loaded from the sequences.
    """

    def __init__(
        self,
        data_loader: DataImporter,
        session_maker: ExperimentSessionMaker,
        *args,
        **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)

        self.setupUi(self)

        self.session_maker = session_maker
        self.path_view = AsyncPathHierarchyView(self.session_maker, self)
        paths_dock = QDockWidget("Sequences", self)
        paths_dock.setWidget(self.path_view)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, paths_dock)
        self.docks_menu.addAction(paths_dock.toggleViewAction())
        self.loader = DataLoader(data_loader, session_maker, self)
        loader_dock = QDockWidget("Watchlist", self)
        loader_dock.setWidget(self.loader)
        self.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, loader_dock)
        self.docks_menu.addAction(loader_dock.toggleViewAction())

        self.path_view.sequence_double_clicked.connect(
            self.loader.add_sequence_to_watchlist
        )
        self.view = ErrorBarView(self)
        self.setCentralWidget(self.view)

    async def start(self):
        async with anyio.create_task_group() as tg:
            tg.start_soon(self.path_view.run_async)
            tg.start_soon(self.loader.process)
            tg.start_soon(self.update_view)

    async def update_view(self):
        while True:
            sequences_data = self.loader.get_sequences_data()
            non_empty_dataframes = [
                d for d in sequences_data.values() if not d.is_empty()
            ]
            if non_empty_dataframes:
                data = await anyio.to_thread.run_sync(
                    polars.concat, non_empty_dataframes
                )
            else:
                data = polars.DataFrame()
            await self.view.update_data(data)
            await anyio.sleep(400e-3)



from typing import Optional, assert_never

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QComboBox, QWidget

from caqtus.device.sequencer import (
    Trigger,
    SoftwareTrigger,
    ExternalClockOnChange,
    ExternalClock,
    ExternalTriggerStart,
)


class TriggerSelector(QComboBox):
    """A widget to select a trigger for a sequencer.

    Signals:
        trigger_changed: Emitted when the trigger is changed, either by the user or
            programmatically.
    """

    trigger_changed = Signal()

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.addItems(
            ["Software", "External start", "External clock", "External adaptive clock"]
        )
        self.setCurrentIndex(0)
        self.currentIndexChanged.connect(self._on_trigger_changed)

    def set_trigger(self, trigger: Trigger) -> None:
        """Set the trigger to be displayed."""

        match trigger:
            case SoftwareTrigger():
                self.setCurrentIndex(0)
            case ExternalTriggerStart():
                self.setCurrentIndex(1)
            case ExternalClock():
                self.setCurrentIndex(2)
            case ExternalClockOnChange():
                self.setCurrentIndex(3)
            case _:
                assert_never(trigger)

    def get_trigger(self) -> Trigger:
        """Get the trigger currently selected."""

        index = self.currentIndex()
        if index == 0:
            return SoftwareTrigger()
        elif index == 1:
            return ExternalTriggerStart()
        elif index == 2:
            return ExternalClock()
        elif index == 3:
            return ExternalClockOnChange()
        else:
            assert False

    def _on_trigger_changed(self, *args, **kwargs) -> None:
        self.trigger_changed.emit()



from PySide6.QtWidgets import QLineEdit


class SerialPortWidget(QLineEdit):
    """A widget to select a serial port"""

    def set_port(self, port: str):
        self.setText(port)

    def get_port(self) -> str:
        return self.text()



from ._experiment_session import SQLExperimentSession
from ._serializer import Serializer
from ._session_maker import (
    SQLExperimentSessionMaker,
    SQLiteExperimentSessionMaker,
    PostgreSQLExperimentSessionMaker,
    PostgreSQLConfig,
)
from ._table_base import create_tables

__all__ = [
    "create_tables",
    "Serializer",
    "SQLExperimentSessionMaker",
    "SQLExperimentSession",
    "SQLiteExperimentSessionMaker",
    "PostgreSQLExperimentSessionMaker",
    "PostgreSQLConfig",
]



import decimal
from abc import ABC, abstractmethod
from typing import (
    Type,
    TypeVar,
    Generic,
)

import attrs

from caqtus.device.configuration import DeviceConfiguration
from .._time_step import TimeStep
from ..channel_commands import ChannelOutput
from ..runtime import Sequencer
from ..trigger import Trigger, is_trigger


@attrs.define
class ChannelConfiguration(ABC):
    """Contains information to computer the output of a channel."""

    description: str = attrs.field(
        converter=str,
        on_setattr=attrs.setters.convert,
    )
    output: ChannelOutput = attrs.field(
        validator=attrs.validators.instance_of(ChannelOutput),
        on_setattr=attrs.setters.validate,
    )


@attrs.define
class DigitalChannelConfiguration(ChannelConfiguration):
    def __str__(self):
        return f"digital channel '{self.description}'"


@attrs.define
class AnalogChannelConfiguration(ChannelConfiguration):
    output_unit: str = attrs.field(
        converter=str,
        on_setattr=attrs.setters.convert,
    )

    def __str__(self):
        return f"analog channel '{self.description}' with unit {self.output_unit}"


def validate_trigger(instance, attribute, value):
    if not is_trigger(value):
        raise TypeError(f"Trigger {value} is not of type Trigger")


SequencerType = TypeVar("SequencerType", bound=Sequencer)


@attrs.define
class SequencerConfiguration(
    DeviceConfiguration[SequencerType], ABC, Generic[SequencerType]
):
    """Holds the static configuration of a sequencer device.

    Fields:
        time_step: The quantization time step used, in nanoseconds.
            The device can only update its output at times that are integer multiples
            of this time step.
            This is a decimal number to allow sub-nanosecond precision without floating
            point errors.
        channels: The configuration of the channels of the device. The length of this
            list must match the number of channels of the device.
        trigger: The trigger that starts the execution of the sequencer.
    """

    time_step: TimeStep = attrs.field(
        converter=decimal.Decimal,
        validator=attrs.validators.gt(decimal.Decimal(0)),
        on_setattr=attrs.setters.pipe(attrs.setters.convert, attrs.setters.validate),
    )
    channels: tuple[ChannelConfiguration, ...] = attrs.field(
        converter=tuple,
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(ChannelConfiguration)
        ),
        on_setattr=attrs.setters.pipe(attrs.setters.convert, attrs.setters.validate),
    )
    trigger: Trigger = attrs.field(validator=validate_trigger)

    @abstractmethod
    def channel_types(self) -> tuple[Type[ChannelConfiguration], ...]:
        """Returns the types of the channels of the device."""

        raise NotImplementedError

    @channels.validator  # type: ignore
    def validate_channels(self, _, channels):
        channel_types = self.channel_types()
        number_channels = len(channel_types)
        if len(channels) != number_channels:
            raise ValueError(
                f"The length of channels ({len(channels)}) doesn't match the number of"
                f" channels {number_channels}"
            )
        for channel, channel_type in zip(channels, channel_types, strict=True):
            if not isinstance(channel, channel_type):
                raise TypeError(
                    f"Channel {channel} is not of the expected type {channel_type}"
                )

    def __getitem__(self, item):
        return self.channels[item]



from collections.abc import Mapping, Iterable, Sequence
from typing import Any, TypeVar, TYPE_CHECKING

import attrs
import tblib.pickling_support
from typing_extensions import deprecated

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.session.shot import TimeLanes, TimeLane
from caqtus.types.variable_name import DottedVariableName
from .lane_compilers.timing import get_step_bounds
from ..formatter import fmt
from ..types.expression import Expression
from ..types.parameter import is_quantity, magnitude_in_unit
from ..types.parameter.analog_value import NotQuantityError
from ..types.recoverable_exceptions import InvalidValueError, EvaluationError
from ..types.units import DimensionalityError
from ..types.units.units import InvalidDimensionalityError

if TYPE_CHECKING:
    from caqtus.shot_compilation import DeviceCompiler

LaneType = TypeVar("LaneType", bound=TimeLane)


@attrs.define(slots=False)
class SequenceContext:
    """Contains information about a sequence being compiled."""

    _device_configurations: Mapping[DeviceName, DeviceConfiguration]
    _time_lanes: TimeLanes

    def get_device_configuration(self, device_name: DeviceName) -> DeviceConfiguration:
        """Returns the configuration for the given device.

        raises:
            KeyError: If no configuration is found for the given device.
        """

        return self._device_configurations[device_name]

    def get_all_device_configurations(self) -> Mapping[DeviceName, DeviceConfiguration]:
        """Returns all device configurations available in this sequence."""

        return self._device_configurations

    def get_lane(self, name: str) -> TimeLane:
        """Returns the time lane with the given name.

        raises:
            KeyError: If no lane with the given name is not found in the sequence
            context.
        """

        return self._time_lanes.lanes[name]

    def get_lanes_with_type(self, lane_type: type[LaneType]) -> Mapping[str, LaneType]:
        """Returns the lanes used during the shot with the given type."""

        return {
            name: lane
            for name, lane in self._time_lanes.lanes.items()
            if isinstance(lane, lane_type)
        }

    def get_step_names(self) -> tuple[str, ...]:
        """Returns the names of the steps in the sequence."""

        return tuple(self._time_lanes.step_names)


@attrs.define
class ShotContext:
    """Contains information about a shot being compiled."""

    _sequence_context: SequenceContext
    _variables: Mapping[DottedVariableName, Any]
    _device_compilers: Mapping[DeviceName, "DeviceCompiler"]

    _step_durations: tuple[float, ...] = attrs.field(init=False)
    _step_bounds: tuple[float, ...] = attrs.field(init=False)
    _was_lane_used: dict[str, bool] = attrs.field(init=False)
    _computed_shot_parameters: dict[DeviceName, Mapping[str, Any]] = attrs.field(
        init=False
    )

    @property
    def _time_lanes(self) -> TimeLanes:
        # noinspection PyProtectedMember
        return self._sequence_context._time_lanes

    def __attrs_post_init__(self):
        self._step_durations = tuple(
            evaluate_step_durations(
                self._time_lanes.step_names,
                self._time_lanes.step_durations,
                self._variables,
            )
        )
        self._step_bounds = tuple(get_step_bounds(self._step_durations))
        self._was_lane_used = {name: False for name in self._time_lanes.lanes}
        self._computed_shot_parameters = {}

    def get_lane(self, name: str) -> TimeLane:
        """Returns the lane with the given name for the shot.

        raises:
            KeyError: If no lane with the given name is present for the shot.
        """

        lane = self._sequence_context.get_lane(name)
        self.mark_lane_used(name)

        return lane

    def mark_lane_used(self, name: str) -> None:
        """Signal that a lane was consumed during the shot.

        Raises:
            KeyError: If no lane with the given name is present for the shot.
        """

        self._was_lane_used[name] = True

    def get_lanes_with_type(self, lane_type: type[LaneType]) -> Mapping[str, LaneType]:
        """Returns the lanes used during the shot with the given type."""

        # Unclear if the lanes obtained here should be marked as used or not.
        return self._sequence_context.get_lanes_with_type(lane_type)

    def get_step_names(self) -> tuple[str, ...]:
        """Returns the names of the steps in the shot."""

        return self._sequence_context.get_step_names()

    def get_step_durations(self) -> tuple[float, ...]:
        """Returns the durations of each step in seconds."""

        return self._step_durations

    def get_step_start_times(self) -> Sequence[float]:
        """Returns the times at which each step starts.

        Returns:
            A sequence representing the start times of each step in seconds.

            For steps with durations [d_0, d_1, ..., d_(n-1)], the returned values are
            [0, d_0, d_0 + d_1, ..., d_0 + ... + d_(n-1)].

            The returned sequence has one more element than the number of steps.
            The last element is the total duration of the shot.
        """

        return self._step_bounds

    @deprecated("Use get_step_start_times instead")
    def get_step_bounds(self) -> Sequence[float]:
        return self.get_step_start_times()

    def get_shot_duration(self) -> float:
        """Returns the total duration of the shot in seconds."""

        return self._step_bounds[-1]

    def get_variables(self) -> Mapping[DottedVariableName, Any]:
        """Returns the variables for the shot."""

        return self._variables

    def get_device_config(self, device_name: DeviceName) -> DeviceConfiguration:
        """Returns the configuration for the given device.

        raises:
            KeyError: If no configuration is found for the given device.
        """

        return self._sequence_context.get_device_configuration(device_name)

    def get_device_compiler(self, device_name: DeviceName) -> "DeviceCompiler":
        """Returns the device compiler for the given device name.

        Raises:
            KeyError: If the requested device is not in use for the current shot.
        """

        return self._device_compilers[device_name]

    def get_shot_parameters(self, device_name: DeviceName) -> Mapping[str, Any]:
        """Returns the parameters computed for the given device."""

        if device_name in self._computed_shot_parameters:
            return self._computed_shot_parameters[device_name]
        else:
            compiler = self._device_compilers[device_name]
            try:
                shot_parameters = compiler.compile_shot_parameters(self)
            except Exception as e:
                raise DeviceCompilationError(
                    fmt(
                        "Couldn't compile parameters for {:device}",
                        device_name,
                    )
                ) from e
            self._computed_shot_parameters[device_name] = shot_parameters
            return shot_parameters

    def _unused_lanes(self) -> set[str]:
        return {name for name, used in self._was_lane_used.items() if not used}


@tblib.pickling_support.install
class DeviceCompilationError(Exception):
    """Raised when compilation for a device fails."""

    pass


def evaluate_step_durations(
    step_names: Iterable[str],
    step_durations: Iterable[Expression],
    variables: Mapping[DottedVariableName, Any],
) -> list[float]:
    result = []

    for step, (name, duration) in enumerate(zip(step_names, step_durations)):
        try:
            evaluated = duration.evaluate(variables)
        except EvaluationError as e:
            raise EvaluationError(
                fmt(
                    "Couldn't evaluate {:expression} for duration of {:step}",
                    duration,
                    (step, name),
                )
            ) from e

        if not is_quantity(evaluated):
            raise NotQuantityError(
                fmt(
                    "{:expression} for duration of {:step} does not evaluate "
                    "to a quantity",
                    duration,
                    (step, name),
                )
            )

        try:
            seconds = magnitude_in_unit(evaluated, "s")
        except DimensionalityError as error:
            raise InvalidDimensionalityError(
                fmt(
                    "Couldn't convert {:expression} for duration of {:step} to "
                    "seconds",
                    duration,
                    (step, name),
                )
            ) from error
        if seconds < 0:
            raise InvalidValueError(
                fmt(
                    "{:expression} for duration of {:step} is negative",
                    duration,
                    (step, name),
                )
            )
        result.append(float(seconds))
    return result



from collections.abc import Mapping
from typing import Any

from caqtus.device import DeviceName, DeviceController
from caqtus.types.data import DataLabel, Data
from . import ShotEventDispatcher
from ._shot_event_dispatcher import DeviceRunConfig
from ...device.remote import DeviceProxy


class ShotRunner:
    def __init__(
        self,
        devices: Mapping[DeviceName, DeviceProxy],
        controller_types: Mapping[DeviceName, type[DeviceController]],
    ):
        if set(devices.keys()) != set(controller_types.keys()):
            raise ValueError("The devices and controller_types must have the same keys")
        self.devices = devices
        self.controller_types = controller_types

    async def run_shot(
        self,
        device_parameters: Mapping[DeviceName, Mapping[str, Any]],
        timeout: float,
    ) -> Mapping[DataLabel, Data]:
        event_dispatcher = ShotEventDispatcher(
            {
                name: DeviceRunConfig(
                    device=self.devices[name],
                    controller_type=self.controller_types[name],
                    parameters=device_parameters[name],
                )
                for name in self.devices
            }
        )
        return await event_dispatcher.run_shot(timeout)



from abc import ABC, abstractmethod
from typing import NewType, Iterable

import attrs
import numpy as np

Width = NewType("Width", int)
Height = NewType("Height", int)


@attrs.define
class ROI(ABC):
    """Base class for regions of interest inside an image.

    Attributes:
        original_image_size: The size of the original image as (width, height)
    """

    original_image_size: tuple[Width, Height] = attrs.field(
        on_setattr=attrs.setters.validate,
    )

    @original_image_size.validator
    def _validate_original_image_size(self, _, value):
        if not isinstance(value, tuple):
            raise TypeError(f"original_image_size must be a tuple, got {type(value)}")
        if len(value) != 2:
            raise ValueError(
                "original_image_size must be a tuple (width, height) with two elements"
            )
        if not all(isinstance(x, int) for x in value):
            raise ValueError("original_image_size must be a tuple of integers")
        if not all(x > 0 for x in value):
            raise ValueError("original_image_size must be a tuple of positive integers")

    def get_mask(self) -> np.ndarray:
        """A boolean array with the same shape as the original image.

        True values indicate that the pixel is part of the region of interest."""

        mask = np.full(self.original_image_size, False)
        mask[*self.get_indices()] = True
        return mask

    @abstractmethod
    def get_indices(self) -> tuple[Iterable[int], Iterable[int]]:
        """Return the indices of the pixels in the ROI."""

        raise NotImplementedError

    @property
    def original_width(self) -> int:
        """Return the width of the original image."""

        return self.original_image_size[0]

    @property
    def original_height(self) -> int:
        """Return the height of the original image."""

        return self.original_image_size[1]



from typing import Optional

from caqtus.device import DeviceName
from caqtus.gui.common.NodeGraphQt import BaseNode


class DeviceTriggerNode(BaseNode):
    __identifier__ = "caqtus.sequencer_node.source"
    NODE_NAME = "Device trigger"

    def __init__(self):
        super().__init__()
        self.add_output("out", multi_output=False, display_name=False)
        self.default_port = self.add_input("absent?", multi_input=False)
        self.add_text_input(
            "Device",
            text="...",
            placeholder_text="device name",
            tooltip="The name of the device for which to generate a trigger",
        )

    def set_device_name(self, device_name: DeviceName) -> None:
        self.set_property("Device", device_name)

    def get_device_name(self) -> DeviceName:
        return DeviceName(str(self.get_property("Device")))

    def get_default_node(self) -> Optional[BaseNode]:
        input_nodes = self.connected_input_nodes()[self.default_port]
        if len(input_nodes) == 0:
            return None
        elif len(input_nodes) == 1:
            return input_nodes[0]
        else:
            assert False, "There can't be multiple nodes connected to the input"



"""This module implement the various components for sequencer devices.
Sequencers are devices that output values at regular time intervals.
"""

from . import channel_commands
from . import instructions
from ._controller import SequencerController
from ._converter import converter
from ._proxy import SequencerProxy
from ._time_step import TimeStep
from .compilation import SequencerCompiler
from .configuration import (
    SequencerConfiguration,
    ChannelConfiguration,
    DigitalChannelConfiguration,
    AnalogChannelConfiguration,
)
from .runtime import Sequencer
from .trigger import (
    Trigger,
    SoftwareTrigger,
    ExternalTriggerStart,
    ExternalClock,
    ExternalClockOnChange,
    TriggerEdge,
)

__all__ = [
    "SequencerConfiguration",
    "Sequencer",
    "Trigger",
    "SoftwareTrigger",
    "ExternalClock",
    "ExternalTriggerStart",
    "ExternalClockOnChange",
    "TriggerEdge",
    "ChannelConfiguration",
    "DigitalChannelConfiguration",
    "AnalogChannelConfiguration",
    "SequencerCompiler",
    "SequencerProxy",
    "SequencerController",
    "channel_commands",
    "instructions",
    "converter",
    "TimeStep",
]



from caqtus.utils import serialization
from .arbitrary_roi import ArbitraryROI
from .rectangular_roi import RectangularROI
from .roi import ROI, Width, Height
from .rotated_rectangular_roi import RotatedRectangularROI

serialization.include_subclasses(ROI)

__all__ = [
    "ArbitraryROI",
    "RectangularROI",
    "RotatedRectangularROI",
    "ROI",
    "Width",
    "Height",
]



import sys

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication, QMainWindow

from spinner import WaitingSpinner

app = QApplication(sys.argv)

window = QMainWindow()
window.show()

spinner = WaitingSpinner(
    window,
    disable_parent_when_spinning=True,
    modality=Qt.WindowModality.ApplicationModal,
)
spinner.start()  # starts spinning
app.exec()



import ast
import re
from collections.abc import Mapping
from functools import cached_property
from typing import Optional, Any

import numpy
import token_utils

from caqtus.utils import serialization
from ..recoverable_exceptions import EvaluationError
from ..units import units
from ..variable_name import DottedVariableName, VariableName

EXPRESSION_REGEX = re.compile(".*")


def square_wave(t, period, duty_cycle=0.5, low=0, high=1):
    x = t / period
    x = x - numpy.floor(x)
    is_high = x < duty_cycle
    result = numpy.where(is_high, high, low)
    return result


#: Built-in functions and constants available in expressions.
BUILTINS: Mapping[str, Any] = {
    "abs": numpy.abs,
    "arccos": numpy.arccos,
    "arcsin": numpy.arcsin,
    "arctan": numpy.arctan,
    "arctan2": numpy.arctan2,
    "ceil": numpy.ceil,
    "cos": numpy.cos,
    "cosh": numpy.cosh,
    "degrees": numpy.degrees,
    "e": numpy.e,
    "exp": numpy.exp,
    "floor": numpy.floor,
    "log": numpy.log,
    "log10": numpy.log10,
    "log2": numpy.log2,
    "pi": numpy.pi,
    "radians": numpy.radians,
    "sin": numpy.sin,
    "sinh": numpy.sinh,
    "sqrt": numpy.sqrt,
    "tan": numpy.tan,
    "tanh": numpy.tanh,
    "square_wave": square_wave,
    "max": max,
    "min": min,
    "Enabled": True,
    "Disabled": False,
} | {str(name): value for name, value in units.items()}


class Expression:
    """Unevaluated expression.

    This class is a wrapper around a python string expression that can be evaluated
    later when the values of the variables it depends on are known.

    If the expression contains syntax errors, they only will be raised when the
    expression is evaluated.

    Args:
        body: the expression body.
            This is a string expression that must follow the specifications of the
            python syntax, with some exceptions.
        implicit_multiplication: if True, the expression will be parsed to allow
            implicit multiplication, even though it is not regular python.
            For example, 'a b' will be parsed as 'a * b'.
            If set to False, a syntax error will be raised in such cases when evaluating
            the expression.
        allow_percentage: if True, the expression will be parsed to understand the
            use of the % symbol as a multiplication by 0.01.
            If set to False, the % symbol will be understood as a modulo operator.
        allow_degree: if True, the expression will be parsed to understand the use
            of the ° symbol as the degree symbol and will be replaced by the name
            `deg` in the expression.
            If set to False, the ° symbol will not be replaced and evaluating the
            expression will raise a SyntaxError.
    """

    def __init__(
        self,
        body: str,
        implicit_multiplication: bool = True,
        allow_percentage: bool = True,
        allow_degree: bool = True,
    ):
        self.body = body
        self._implicit_multiplication = implicit_multiplication
        self._allow_percentage = allow_percentage
        self._allow_degree = allow_degree

    @property
    def body(self) -> str:
        return self._body

    # noinspection PyPropertyAccess
    @body.setter
    def body(self, value):
        if not isinstance(value, str):
            raise TypeError(f"Expression body must be a string, got {value}")
        self._body = value
        if hasattr(self, "ast"):
            del self.ast
        if hasattr(self, "code"):
            del self.code
        if hasattr(self, "upstream_variables"):
            del self.upstream_variables

    def __repr__(self) -> str:
        return f"Expression('{self.body}')"

    def __str__(self) -> str:
        return self.body

    def check_syntax(self) -> Optional[SyntaxError]:
        """Force parsing of the expression.

        It is not necessary to call this method explicitly, as the expression will be
        parsed automatically when it is evaluated. However, this method can be used
        to force the parsing to happen at a specific time, for example to catch
        syntax errors early.

        Returns:
            None if the expression is valid, or a SyntaxError otherwise.
        """

        try:
            self._parse_ast()
        except SyntaxError as error:
            return error

    def evaluate(self, variables: Mapping[DottedVariableName, Any]) -> Any:
        """Evaluate an expression on specific values for its variables.

        Args:
            variables: a mapping of variable names to their values.
                The variables the expression depends upon will be replaced by the values
                in this mapping.
                The mapping is also concatenated with the :data:`BUILTINS` dictionary.

        Raises:
            EvaluationError: if an error occurred during evaluation.
        """

        return self._evaluate({str(expr): variables[expr] for expr in variables})

    def _evaluate(self, variables: Mapping[str, Any]) -> Any:
        try:
            value = eval(self.code, {"__builtins__": BUILTINS}, variables)
        except Exception as error:
            raise EvaluationError(f"Could not evaluate <{self.body}>") from error
        return value

    @cached_property
    def upstream_variables(self) -> frozenset[VariableName]:
        """Return the name of the other variables the expression depend on."""

        variables = set()

        class FindNameVisitor(ast.NodeVisitor):
            def visit_Name(self, node: ast.Name):
                if isinstance(node.ctx, ast.Load):
                    if node.id not in BUILTINS:
                        variables.add(VariableName(node.id))

        FindNameVisitor().visit(self._ast)
        return frozenset(variables)

    @cached_property
    def _ast(self) -> ast.Expression:
        """Computes the abstract syntax tree for this expression"""

        return self._parse_ast()

    def _parse_ast(self) -> ast.Expression:
        expr = self.body
        if self._allow_percentage:
            expr = expr.replace("%", "*(1e-2)")

        if self._allow_degree:
            expr = expr.replace("°", "*deg")

        if self._implicit_multiplication:
            expr = add_implicit_multiplication(expr)

        return ast.parse(expr, mode="eval")

    @cached_property
    def code(self):
        return compile(self._ast, filename="<string>", mode="eval")

    def __eq__(self, other):
        if isinstance(other, Expression):
            return self.body == other.body
        else:
            return NotImplemented

    def __getstate__(self):
        return {
            "body": self.body,
            "implicit_multiplication": self._implicit_multiplication,
            "allow_percentage": self._allow_percentage,
            "allow_degree": self._allow_degree,
        }

    def __setstate__(self, state):
        self.__init__(**state)


serialization.register_unstructure_hook(Expression, lambda expr: expr.body)
serialization.register_structure_hook(Expression, lambda body, _: Expression(body))


def add_implicit_multiplication(source: str) -> str:
    """This adds a multiplication symbol where it would be understood as
    being implicit by the normal way algebraic equations are written but would
    be a SyntaxError in Python. Thus we have::

        2n -> 2*n
        n 2 -> n* 2
        2(a+b) -> 2*(a+b)
        (a+b)2 -> (a+b)*2
        2 3 -> 2* 3
        m n -> m* n
        (a+b)c -> (a+b)*c

    The obvious one (in algebra) being left out is something like ``n(...)``
    which is a function call - and thus valid Python syntax.
    """

    tokens = token_utils.tokenize(source)
    if not tokens:
        return tokens

    prev_token = tokens[0]
    new_tokens = [prev_token]

    for token in tokens[1:]:
        if (
            (
                prev_token.is_number()
                and (token.is_identifier() or token.is_number() or token == "(")
            )
            or (
                prev_token.is_identifier()
                and (token.is_identifier() or token.is_number())
            )
            or (prev_token == ")" and (token.is_identifier() or token.is_number()))
        ):
            new_tokens.append("*")
        new_tokens.append(token)
        prev_token = token

    return token_utils.untokenize(new_tokens)



import contextlib
import operator
import pickle
from collections.abc import Callable, AsyncGenerator, Iterator, AsyncIterator
from typing import Any, TypeVar, TypeAlias, Literal, LiteralString

import anyio
import attrs
import eliot
from anyio.streams.buffered import BufferedByteReceiveStream

from ._prefix_size import receive_with_size_prefix, send_with_size_prefix
from ._server import (
    CallRequest,
    ReturnValue,
    CallResponse,
    CallResponseSuccess,
    CallResponseFailure,
    DeleteProxyRequest,
    TerminateRequest,
    RemoteError,
)
from .proxy import Proxy

T = TypeVar("T")

ReturnedType: TypeAlias = Literal["copy", "proxy"]


@attrs.frozen
class MethodCaller:
    method: LiteralString

    def __call__(self, obj: Any, *args: Any, **kwargs: Any) -> Any:
        return getattr(obj, self.method)(*args, **kwargs)

    def __str__(self):
        return f"call method {self.method}"


class RPCClient:
    def __init__(self, host: str, port: int):
        self._host = host
        self._port = port

        self._exit_stack = contextlib.AsyncExitStack()

    async def __aenter__(self):
        await self._exit_stack.__aenter__()
        self._exit_stack.enter_context(eliot.start_action(action_type="rpc client"))
        self._stream = await anyio.connect_tcp(self._host, self._port)
        self._receive_stream = BufferedByteReceiveStream(self._stream)
        await self._exit_stack.enter_async_context(self._stream)
        self._exit_stack.push_async_callback(self.terminate)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self._exit_stack.__aexit__(exc_type, exc_value, traceback)

    async def call(
        self,
        fun: Callable[..., T],
        *args: Any,
        **kwargs: Any,
    ) -> T:
        # We cancel this scope to be sure that the request always come to completion.
        # The issue is if we make a request and an external exceptions cancel waiting for the answer, then we would not
        # read the answer, and it would remain in the buffer.
        with anyio.CancelScope(shield=True):
            with eliot.start_action(action_type="call", function=str(fun)):
                with eliot.start_action(action_type="send"):
                    request = self._build_request(fun, args, kwargs, "copy")
                    pickled = pickle.dumps(request)
                    await send_with_size_prefix(self._stream, pickled)
                with eliot.start_action(action_type="receive"):
                    bytes_response = await receive_with_size_prefix(
                        self._receive_stream
                    )
                    response = pickle.loads(bytes_response)
                    return self._build_result(response)

    async def call_method(
        self,
        obj: Any,
        method: LiteralString,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        return await self.call(
            MethodCaller(method=method),
            obj,
            *args,
            **kwargs,
        )

    async def terminate(self):
        request = TerminateRequest()
        pickled_request = pickle.dumps(request)
        await send_with_size_prefix(self._stream, pickled_request)

    @contextlib.asynccontextmanager
    async def call_method_proxy_result(
        self,
        obj: Any,
        method: LiteralString,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        caller = operator.methodcaller(method, *args, **kwargs)
        async with self.call_proxy_result(caller, obj) as result:
            yield result

    async def get_attribute(self, obj: Any, attribute: LiteralString) -> Any:
        caller = operator.attrgetter(attribute)
        return await self.call(caller, obj)

    @contextlib.asynccontextmanager
    async def call_proxy_result(
        self, fun: Callable[..., T], *args: Any, **kwargs: Any
    ) -> AsyncGenerator[Proxy[T], None]:
        request = self._build_request(fun, args, kwargs, "proxy")
        pickled_request = pickle.dumps(request)
        with anyio.CancelScope(shield=True):
            await send_with_size_prefix(self._stream, pickled_request)
            pickled_response = await receive_with_size_prefix(self._receive_stream)
            response = pickle.loads(pickled_response)

            proxy = self._build_result(response)
            assert isinstance(proxy, Proxy)
            try:
                yield proxy
            finally:
                await self._close_proxy(proxy)

    @contextlib.asynccontextmanager
    async def async_context_manager(
        self, cm_proxy: Proxy[contextlib.AbstractContextManager[T]]
    ) -> AsyncGenerator[Proxy[T], None]:
        with anyio.CancelScope(shield=True):
            stack = contextlib.AsyncExitStack()
            try:
                result_proxy = await stack.enter_async_context(
                    self.call_method_proxy_result(cm_proxy, "__enter__")
                )
            except Exception:
                raise
            else:
                try:
                    async with stack:
                        yield result_proxy
                finally:
                    await self.call_method(cm_proxy, "__exit__", None, None, None)

    async def async_iterator(self, proxy: Proxy[Iterator[T]]) -> AsyncIterator[T]:
        while True:
            try:
                value = await self.call_method(proxy, "__next__")
                yield value
            except RemoteError as error:
                if isinstance(error.__cause__, StopIteration):
                    break
                else:
                    raise

    async def _close_proxy(self, proxy: Proxy[T]) -> None:
        request = DeleteProxyRequest(proxy)
        pickled_request = pickle.dumps(request)
        await send_with_size_prefix(self._stream, pickled_request)

    @staticmethod
    def _build_request(
        fun: Callable[..., T], args: Any, kwargs: Any, returned_value: ReturnedType
    ) -> CallRequest:
        return CallRequest(
            function=fun,
            args=args,
            kwargs=kwargs,
            return_value=(
                ReturnValue.SERIALIZED
                if returned_value == "copy"
                else ReturnValue.PROXY
            ),
        )

    @staticmethod
    def _build_result(response: CallResponse) -> Any:
        match response:
            case CallResponseSuccess(result=result):
                return result
            case CallResponseFailure(error=error):
                raise error



import contextlib
import functools
import logging

from typing import Callable, TypeVar, ParamSpec, Optional

_P = ParamSpec("_P")
_T = TypeVar("_T")


def log_cm_decorator(logger: logging.Logger, level: int = logging.DEBUG):
    """Decorator to log the start and end of a context manager."""

    def decorator(
        func: Callable[_P, contextlib.AbstractContextManager[_T]]
    ) -> Callable[_P, contextlib.AbstractContextManager[_T]]:
        @functools.wraps(func)
        @contextlib.contextmanager
        def wrapper(*args: _P.args, **kwargs: _P.kwargs):
            logger.log(level, f"Enter cm: %s.", func.__name__)
            with func(*args, **kwargs) as cm:
                yield cm
            logger.log(level, f"Exit  cm: %s.", func.__name__)

        return wrapper

    return decorator


def log_async_cm_decorator(logger: logging.Logger, level: int = logging.DEBUG):
    """Decorator to log the start and end of an asynchronous context manager."""

    def decorator(
        func: Callable[_P, contextlib.AbstractAsyncContextManager[_T]]
    ) -> Callable[_P, contextlib.AbstractAsyncContextManager[_T]]:
        @functools.wraps(func)
        @contextlib.asynccontextmanager
        async def wrapper(*args: _P.args, **kwargs: _P.kwargs):
            logger.log(level, f"Enter cm: %s.", func.__name__)
            async with func(*args, **kwargs) as cm:
                yield cm
            logger.log(level, f"Exit  cm: %s.", func.__name__)

        return wrapper

    return decorator


@contextlib.asynccontextmanager
async def log_async_cm(
    cm: contextlib.AbstractAsyncContextManager[_T],
    logger: logging.Logger,
    name: Optional[str] = None,
    level: int = logging.DEBUG,
):
    """Transformer to log the start and end of an asynchronous context manager."""

    to_print = cm if name is None else name
    logger.log(level, "Enter cm: %s.", to_print)
    try:
        async with cm as value:
            yield value
    finally:
        logger.log(level, f"Exit  cm: %s.", to_print)



import contextlib
import functools
import itertools
import logging
import os
import pickle
import warnings
from collections.abc import Callable
from enum import Enum, auto
from typing import Never, Any, TypeVar, Self, ParamSpec

import anyio
import anyio.abc
import anyio.to_thread
import attrs
import tblib.pickling_support
from anyio.streams.buffered import BufferedByteReceiveStream

from ._configuration import (
    RPCConfiguration,
)
from ._prefix_size import receive_with_size_prefix, send_with_size_prefix
from .proxy import Proxy

logger = logging.getLogger(__name__)


@attrs.define
class ObjectReference:
    obj: object
    number_proxies: int


class ReturnValue(Enum):
    SERIALIZED = auto()
    PROXY = auto()


@attrs.define
class CallRequest:
    function: Callable[..., Any]
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    return_value: ReturnValue


@attrs.define
class DeleteProxyRequest:
    proxy: Proxy


@attrs.define
class CallResponseFailure:
    error: Exception


@attrs.define
class CallResponseSuccess:
    result: Any


class TerminateRequest:
    pass


CallResponse = CallResponseFailure | CallResponseSuccess

T = TypeVar("T")


class RPCServer:
    def __init__(self, port: int):
        self._port = port
        self._objects: dict[int, ObjectReference] = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._objects:
            warnings.warn(
                f"Not all objects were properly deleted: {self._objects}.\n"
                f"If you acquired any proxies, make sure to close them."
            )
            self._objects.clear()

    async def run_async(self) -> Never:
        listener = await anyio.create_tcp_listener(local_port=self._port)
        async with contextlib.aclosing(listener):
            await listener.serve(self.handle)

    async def handle(self, client: anyio.abc.ByteStream) -> None:
        async with client:
            receive_stream = BufferedByteReceiveStream(client)
            for _ in itertools.count():
                request_bytes = await receive_with_size_prefix(receive_stream)
                request = pickle.loads(request_bytes)
                if isinstance(request, CallRequest):
                    await self.handle_call_request(client, request)
                elif isinstance(request, DeleteProxyRequest):
                    self.handle_delete_proxy_request(request)
                elif isinstance(request, TerminateRequest):
                    break
                else:
                    raise ValueError(f"Unknown request type: {request}")

    async def handle_call_request(self, client, request: CallRequest) -> None:
        try:
            args = [self.resolve(arg) for arg in request.args]
            kwargs = {key: self.resolve(value) for key, value in request.kwargs.items()}
            # We can't have the function raise StopIteration since the coroutine will
            # then raise this exception which is invalid.
            # To prevent this, we replace StopIteration with our own exception.
            fun = _transform_stop_iteration(request.function)
            value = await anyio.to_thread.run_sync(
                functools.partial(fun, *args, **kwargs)
            )

            if request.return_value == ReturnValue.SERIALIZED:
                result = value
            elif request.return_value == ReturnValue.PROXY:
                result = self.create_proxy(value)
            else:
                assert False, f"Unknown return value: {request.return_value}"
        except _StopIteration:
            # Don't log this exception, as it can occur during normal operation if we're
            # calling __next__ on an iterator.
            await self.send_failure_response(client, request, StopIteration())
        except Exception as e:
            logger.exception(f"Error during request call {request!r}")
            await self.send_failure_response(client, request, e)
        else:
            await self.send_success_response(client, request, result)

    def handle_delete_proxy_request(self, request: DeleteProxyRequest) -> None:
        proxy = request.proxy
        if proxy._pid != os.getpid():
            raise RuntimeError(
                "Proxy cannot be deleted in a different process than the one it was "
                "created in"
            )
        self._objects[proxy._obj_id].number_proxies -= 1
        if self._objects[proxy._obj_id].number_proxies <= 0:
            del self._objects[proxy._obj_id]

    @staticmethod
    async def send_failure_response(
        client: anyio.abc.ByteStream, request: CallRequest, e: Exception
    ) -> None:
        try:
            raise RemoteCallError(f"Error during call to {request.function}") from e
        except RemoteCallError as error:
            response = CallResponseFailure(error=error)
        pickled_response = pickle.dumps(response)
        await send_with_size_prefix(client, pickled_response)

    @staticmethod
    async def send_success_response(
        client: anyio.abc.ByteStream, request: CallRequest, result: Any
    ) -> None:
        response = CallResponseSuccess(result=result)
        pickled_response = pickle.dumps(response)
        await send_with_size_prefix(client, pickled_response)

    def create_proxy(self, obj: T) -> Proxy[T]:
        obj_id = id(obj)
        if obj_id not in self._objects:
            self._objects[obj_id] = ObjectReference(obj=obj, number_proxies=0)
        proxy = Proxy(os.getpid(), obj_id)
        self._objects[obj_id].number_proxies += 1
        return proxy

    def get_referent(self, proxy: Proxy[T]) -> T:
        if proxy._pid != os.getpid():
            raise InvalidProxyError(
                "Proxy cannot be resolved in a different process than the one it was "
                "created in"
            )
        try:
            return self._objects[proxy._obj_id].obj
        except KeyError as e:
            raise InvalidProxyError(
                f"{proxy} is referring to an object that does not exist on the "
                f"server"
            ) from e

    def resolve(self, obj: Proxy[T] | T) -> T:
        if isinstance(obj, Proxy):
            return self.get_referent(obj)
        else:
            return obj

    def run(self) -> Never:
        anyio.run(self.run_async, backend="trio")


class Server:
    def __init__(self, config: RPCConfiguration) -> None:
        self._server = RPCServer(config.port)

    def wait_for_termination(self) -> None:
        self._server.run()

    def __enter__(self) -> Self:
        self._server.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._server.__exit__(exc_type, exc_value, traceback)
        logger.info("Server stopped")


@tblib.pickling_support.install
class RemoteError(Exception):
    """Base class for errors that occur on the server side."""

    pass


@tblib.pickling_support.install
class RemoteCallError(RemoteError):
    """Error that occurs when calling a remote function."""

    pass


@tblib.pickling_support.install
class InvalidProxyError(RemoteError):
    pass


P = ParamSpec("P")


def _transform_stop_iteration(fun: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(fun)
    def wrapper(*args: Any, **kwargs: Any) -> T:
        try:
            return fun(*args, **kwargs)
        except StopIteration:
            raise _StopIteration() from None

    return wrapper


class _StopIteration(BaseException):
    pass



from numbers import Real
from typing import Any, Optional, TypeGuard, overload, TypeVar

import numpy as np
from numpy.typing import NDArray

from ..recoverable_exceptions import InvalidTypeError
from ..units import Quantity, Unit, dimensionless
from ..units.units import UnitLike

AnalogValue = Real | NDArray[np.floating] | Quantity


class NotAnalogValueError(InvalidTypeError):
    pass


class NotQuantityError(InvalidTypeError):
    pass


def is_analog_value(value: Any) -> TypeGuard[AnalogValue]:
    """Returns True if the value is an analog value, False otherwise."""

    if isinstance(value, np.ndarray):
        return issubclass(value.dtype.type, np.floating)
    elif isinstance(value, (Real, Quantity)):
        return True
    return False


def is_quantity(value: Any) -> TypeGuard[Quantity]:
    """Returns True if the value is a quantity, False otherwise."""

    return isinstance(value, Quantity)


def get_unit(value: AnalogValue) -> Optional[Unit]:
    """Returns the unit of the value if it has one, None otherwise."""

    if isinstance(value, Quantity):
        return value.units
    return None


_R = TypeVar("_R", bound=Real)


@overload
def magnitude_in_unit(value: _R, unit: None) -> _R: ...


@overload
def magnitude_in_unit(value: _R, unit: UnitLike) -> _R: ...


@overload
def magnitude_in_unit(
    value: NDArray[np.floating], unit: None
) -> NDArray[np.floating]: ...


@overload
def magnitude_in_unit(
    value: Quantity, unit: UnitLike
) -> Real | NDArray[np.floating]: ...


def magnitude_in_unit(value, unit):
    """Return the magnitude of a value in the given unit."""

    if is_quantity(value):
        if unit is None:
            return value.to(dimensionless).magnitude
        return value.to(unit).magnitude
    else:
        if unit is None:
            return value
        else:
            value = value * dimensionless
            return value.to(unit).magnitude


def add_unit(
    magnitude: Real | NDArray[np.floating], unit: Optional[Unit]
) -> AnalogValue:
    """Add a unit to a magnitude."""

    if unit is None:
        return magnitude
    return Quantity(magnitude, unit)


def are_units_compatible(unit1: Optional[Unit], unit2: Optional[Unit]) -> bool:
    """Return True if the two units are compatible, False otherwise."""

    if unit1 is None:
        return unit2 is None
    if unit2 is None:
        return unit1 is None

    return unit1.is_compatible_with(unit2)



from __future__ import annotations

import abc
import datetime
from typing import Optional, TypeGuard

import anyio
import attrs
from PySide6.QtCore import QObject, QAbstractItemModel, QModelIndex, Qt, QDateTime
from PySide6.QtGui import QStandardItemModel, QStandardItem

from caqtus.session import (
    ExperimentSessionMaker,
    PureSequencePath,
    ExperimentSession,
    AsyncExperimentSession,
)
from caqtus.session._return_or_raise import unwrap
from caqtus.session.path_hierarchy import PathNotFoundError
from caqtus.session.sequence import State
from caqtus.session.sequence_collection import (
    PathIsSequenceError,
    SequenceStats,
    PathIsNotSequenceError,
)
from caqtus.types.iteration import is_unknown

NODE_DATA_ROLE = Qt.UserRole + 1


def get_item_data(item: QStandardItem) -> Node:
    data = item.data(NODE_DATA_ROLE)
    assert is_node(data)
    return data


@attrs.define
class FolderNode:
    path: PureSequencePath
    has_fetched_children: bool
    creation_date: datetime.datetime


@attrs.define
class SequenceNode(abc.ABC):
    path: PureSequencePath
    stats: SequenceStats
    creation_date: datetime.datetime
    last_query_time: datetime.datetime


Node = FolderNode | SequenceNode


def is_node(value) -> TypeGuard[Node]:
    return isinstance(value, (FolderNode, SequenceNode))


class AsyncPathHierarchyModel(QAbstractItemModel):
    def __init__(
        self, session_maker: ExperimentSessionMaker, parent: Optional[QObject] = None
    ):
        super().__init__(parent)
        self.session_maker = session_maker

        self.tree = QStandardItemModel(self)
        self.tree.invisibleRootItem().setData(
            FolderNode(
                path=PureSequencePath.root(),
                has_fetched_children=False,
                creation_date=datetime.datetime.min,
            ),
            NODE_DATA_ROLE,
        )

    def index(self, row, column, parent=QModelIndex()):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()
        parent_item = (
            parent.internalPointer()
            if parent.isValid()
            else self.tree.invisibleRootItem()
        )
        child_item = parent_item.child(row)
        return (
            self.createIndex(row, column, child_item) if child_item else QModelIndex()
        )

    def parent(self, index=QModelIndex()):
        if not index.isValid():
            return QModelIndex()
        child_item = index.internalPointer()
        parent_item = child_item.parent()
        if parent_item is None:
            return QModelIndex()
        return (
            self.createIndex(parent_item.row(), 0, parent_item)
            if parent_item is not self.tree.invisibleRootItem()
            else QModelIndex()
        )

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        item = self._get_item(index)
        node_data = get_item_data(item)
        flags = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        if isinstance(node_data, SequenceNode):
            flags |= Qt.ItemFlag.ItemNeverHasChildren
        return flags

    def get_path(self, index: QModelIndex) -> PureSequencePath:
        item = self._get_item(index)
        node_data = get_item_data(item)
        return node_data.path

    def _get_item(self, index) -> QStandardItem:
        result = (
            index.internalPointer()
            if index.isValid()
            else self.tree.invisibleRootItem()
        )
        assert isinstance(result, QStandardItem)
        return result

    def rowCount(self, parent=QModelIndex()):
        if parent.column() > 0:
            return 0
        parent_item = self._get_item(parent)
        node_data = get_item_data(parent_item)
        match node_data:
            case SequenceNode():
                return 0
            case FolderNode(has_fetched_children=True):
                return parent_item.rowCount()
            case FolderNode(has_fetched_children=False):
                return 0

    def hasChildren(self, parent=QModelIndex()) -> bool:
        parent_item = self._get_item(parent)
        node_data = get_item_data(parent_item)
        match node_data:
            case SequenceNode():
                return False
            case FolderNode(has_fetched_children=True):
                return parent_item.rowCount() > 0
            case FolderNode(has_fetched_children=False):
                return True

    def canFetchMore(self, parent) -> bool:
        parent_item = self._get_item(parent)
        node_data = get_item_data(parent_item)
        match node_data:
            case SequenceNode():
                return False
            case FolderNode(has_fetched_children=already_fetched):
                return not already_fetched

    def fetchMore(self, parent):
        parent_item = self._get_item(parent)
        node_data = get_item_data(parent_item)
        match node_data:
            case SequenceNode():
                return
            case FolderNode(has_fetched_children=True):
                return
            case FolderNode(path=parent_path, has_fetched_children=False):
                assert parent_item.rowCount() == 0
                with self.session_maker() as session:
                    children_result = session.paths.get_children(parent_path)
                    try:
                        children = unwrap(children_result)
                    except PathIsSequenceError:
                        self.handle_folder_became_sequence(parent, session)
                        return
                    except PathNotFoundError:
                        self.handle_path_was_deleted(parent)
                        return
                    self.beginInsertRows(parent, 0, len(children) - 1)
                    for child_path in children:
                        child_item = self._build_item(child_path, session)
                        parent_item.appendRow(child_item)
                    node_data.has_fetched_children = True
                    self.endInsertRows()

    @staticmethod
    def _build_item(
        path: PureSequencePath, session: ExperimentSession
    ) -> QStandardItem:
        assert session.paths.does_path_exists(path)
        item = QStandardItem()
        item.setData(path.name, Qt.DisplayRole)
        is_sequence = unwrap(session.sequences.is_sequence(path))
        creation_date = unwrap(session.paths.get_path_creation_date(path))
        if is_sequence:
            stats = unwrap(session.sequences.get_stats(path))
            item.setData(
                SequenceNode(
                    path=path,
                    stats=stats,
                    creation_date=creation_date,
                    last_query_time=get_update_date(),
                ),
                NODE_DATA_ROLE,
            )
        else:
            item.setData(
                FolderNode(
                    path=path, has_fetched_children=False, creation_date=creation_date
                ),
                NODE_DATA_ROLE,
            )
        return item

    @staticmethod
    async def _build_item_async(
        path: PureSequencePath, session: AsyncExperimentSession
    ) -> QStandardItem:
        assert await session.paths.does_path_exists(path)
        item = QStandardItem()
        item.setData(path.name, Qt.DisplayRole)
        is_sequence = unwrap(await session.sequences.is_sequence(path))
        creation_date = unwrap(await session.paths.get_path_creation_date(path))
        if is_sequence:
            stats = unwrap(await session.sequences.get_stats(path))
            item.setData(
                SequenceNode(
                    path=path,
                    stats=stats,
                    creation_date=creation_date,
                    last_query_time=get_update_date(),
                ),
                NODE_DATA_ROLE,
            )
        else:
            item.setData(
                FolderNode(
                    path=path, has_fetched_children=False, creation_date=creation_date
                ),
                NODE_DATA_ROLE,
            )
        return item

    def columnCount(self, parent=QModelIndex()):
        return 5

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                if section == 0:
                    return "Name"
                elif section == 1:
                    return "Status"
                elif section == 2:
                    return "Progress"
                elif section == 3:
                    return "Duration"
                elif section == 4:
                    return "Date created"
            else:
                return section
        return None

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        """Get the data for a specific index in the model.

        The displayed data returned for each column is as follows:
        0: Name
        A string with the name of the folder or sequence.
        1: Status
        The status of the sequence.
        It is None for folders and a SequenceStats object for sequences.
        2: Progress
        A string representing the number of completed shots and the total
        number of shots of the sequence.
        It is None for folders.
        3: Duration
        A string representing the elapsed and remaining time of the
        sequence.
        It is None for folders.
        4: Date created
        A QDateTime object representing the date and time when the
        sequence or folder was created.
        """

        if not index.isValid() or role != Qt.ItemDataRole.DisplayRole:
            return None
        item = self._get_item(index)
        node_data = get_item_data(item)
        if index.column() == 0:
            return node_data.path.name
        elif index.column() == 1:
            if isinstance(node_data, SequenceNode):
                return node_data.stats
            else:
                return None
        elif index.column() == 2:
            if isinstance(node_data, SequenceNode):
                return (
                    f"{node_data.stats.number_completed_shots}"
                    f"/{node_data.stats.expected_number_shots}"
                )
            else:
                return None
        elif index.column() == 3:
            if isinstance(node_data, SequenceNode):
                return format_duration(node_data.stats, node_data.last_query_time)
            else:
                return None
        elif index.column() == 4:
            return QDateTime(node_data.creation_date.astimezone(None))  # type: ignore

    async def watch_session(self) -> None:
        while True:
            await self.update_from_session()
            await anyio.sleep(0)

    async def update_from_session(self) -> None:
        await self.prune()
        await self.add_new_paths()
        for row in range(self.rowCount()):
            await self.update_stats(self.index(row, 0))

    async def update_stats(self, index: QModelIndex) -> None:
        """Update the stats of sequences and folders in the model from the session."""

        if not index.isValid():
            # This situation occurs sometimes, but unsure why.
            # Maybe if fetch is called in the middle of an async call?
            return None

        item = self._get_item(index)
        data = get_item_data(item)
        change_detected = False
        async with self.session_maker.async_session() as session:
            creation_date_result = await session.paths.get_path_creation_date(data.path)
            try:
                creation_date = unwrap(creation_date_result)
            except PathNotFoundError:
                self.handle_path_was_deleted(index)
                return
            if creation_date != data.creation_date:
                data.creation_date = creation_date
                change_detected = True
            if isinstance(data, SequenceNode):
                sequence_stats_result = await session.sequences.get_stats(data.path)
                try:
                    stats = unwrap(sequence_stats_result)
                except PathIsNotSequenceError:
                    await self.handle_sequence_became_folder(index, session)
                    return
                if stats != data.stats:
                    data.stats = stats
                    data.last_query_time = get_update_date()
                    change_detected = True
        if change_detected:
            top_left = index.siblingAtColumn(0)
            bottom_right = index.siblingAtColumn(self.columnCount() - 1)
            self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.DisplayRole])
        if isinstance(data, FolderNode):
            for row in range(item.rowCount()):
                await self.update_stats(self.index(row, 0, index))

    async def prune(self, parent: QModelIndex = QModelIndex()) -> None:
        """Removes items from the model that no longer exist in the session."""

        parent_item = self._get_item(parent)
        parent_data = get_item_data(parent_item)
        async with self.session_maker.async_session() as session:
            children_result = await session.paths.get_children(parent_data.path)
            try:
                child_paths = unwrap(children_result)
            except PathIsSequenceError:
                await self.handle_folder_became_sequence_async(parent, session)
                return
            except PathNotFoundError:
                self.handle_path_was_deleted(parent)
                return

        # Need to remove children in reverse order to avoid invalidating rows
        for row in reversed(range(self.rowCount(parent))):
            child = self.index(row, 0, parent)
            child_item = self._get_item(child)
            child_path = get_item_data(child_item).path
            if child_path not in child_paths:
                self.beginRemoveRows(parent, row, row)
                parent_item.removeRow(row)
                self.endRemoveRows()
            else:
                await self.prune(child)

    async def add_new_paths(self, parent: QModelIndex = QModelIndex()) -> None:
        """Add new paths to the model that have been added to the session."""

        parent_item = self._get_item(parent)
        parent_data = get_item_data(parent_item)
        match parent_data:
            case SequenceNode():
                return
            case FolderNode(has_fetched_children=False):
                return
            case FolderNode(path=parent_path, has_fetched_children=True):
                async with self.session_maker.async_session() as session:
                    children_result = await session.paths.get_children(parent_path)
                    try:
                        child_paths = unwrap(children_result)
                    except PathIsSequenceError:
                        await self.handle_folder_became_sequence_async(parent, session)
                        return
                    except PathNotFoundError:
                        self.handle_path_was_deleted(parent)
                        return
                    already_added_paths = {
                        get_item_data(parent_item.child(row)).path
                        for row in range(parent_item.rowCount())
                    }
                    new_paths = child_paths - already_added_paths
                    self.beginInsertRows(
                        parent,
                        parent_item.rowCount(),
                        parent_item.rowCount() + len(new_paths) - 1,
                    )
                    for child_path in new_paths:
                        child_item = await self._build_item_async(child_path, session)
                        parent_item.appendRow(child_item)
                    self.endInsertRows()
                for row in range(self.rowCount(parent)):
                    await self.add_new_paths(self.index(row, 0, parent))

    def handle_folder_became_sequence(
        self, index: QModelIndex, session: ExperimentSession
    ):
        item = self._get_item(index)
        data = get_item_data(item)
        stats = unwrap(session.sequences.get_stats(data.path))
        creation_date = unwrap(session.paths.get_path_creation_date(data.path))
        self.beginRemoveRows(index, 0, item.rowCount() - 1)
        item.setData(
            SequenceNode(
                path=data.path,
                stats=stats,
                creation_date=creation_date,
                last_query_time=get_update_date(),
            ),
            NODE_DATA_ROLE,
        )
        item.removeRows(0, item.rowCount())
        self.endRemoveRows()
        self.emit_index_updated(index)

    async def handle_folder_became_sequence_async(
        self, index: QModelIndex, session: AsyncExperimentSession
    ):
        item = self._get_item(index)
        data = get_item_data(item)
        stats = unwrap(await session.sequences.get_stats(data.path))
        creation_date = unwrap(await session.paths.get_path_creation_date(data.path))
        self.beginRemoveRows(index, 0, item.rowCount() - 1)
        item.setData(
            SequenceNode(
                path=data.path,
                stats=stats,
                creation_date=creation_date,
                last_query_time=get_update_date(),
            ),
            NODE_DATA_ROLE,
        )
        item.removeRows(0, item.rowCount())
        self.endRemoveRows()
        self.emit_index_updated(index)

    def handle_path_was_deleted(self, index: QModelIndex):
        parent = self.parent(index)
        grandparent_item = self._get_item(parent)
        self.beginRemoveRows(parent, index.row(), index.row())
        grandparent_item.removeRow(index.row())
        self.endRemoveRows()

    async def handle_sequence_became_folder(
        self, index: QModelIndex, session: AsyncExperimentSession
    ):
        item = self._get_item(index)
        data = get_item_data(item)
        creation_date = unwrap(await session.paths.get_path_creation_date(data.path))
        assert item.rowCount() == 0
        item.setData(
            FolderNode(
                path=data.path, has_fetched_children=False, creation_date=creation_date
            ),
            NODE_DATA_ROLE,
        )
        self.emit_index_updated(index)

    def emit_index_updated(self, index: QModelIndex) -> None:
        self.dataChanged.emit(
            index.siblingAtColumn(0),
            index.siblingAtColumn(self.columnCount() - 1),
            [Qt.ItemDataRole.DisplayRole],
        )


def format_duration(stats: SequenceStats, updated_time: datetime.datetime) -> str:
    if stats.state == State.DRAFT or stats.state == State.PREPARING:
        return f"--/--"
    elif stats.state == State.RUNNING:
        running_duration = updated_time - stats.start_time
        expected_num_shots = stats.expected_number_shots
        if is_unknown(expected_num_shots) or stats.number_completed_shots == 0:
            remaining = "--"
        else:
            remaining = (
                running_duration
                / stats.number_completed_shots
                * (expected_num_shots - stats.number_completed_shots)
            )
        if isinstance(remaining, datetime.timedelta):
            total = remaining + running_duration
            remaining = _format_seconds(total.total_seconds())
        running_duration = _format_seconds(running_duration.total_seconds())
        return f"{running_duration}/{remaining}"
    elif (
        stats.state == State.FINISHED
        or stats.state == State.CRASHED
        or stats.state == State.INTERRUPTED
    ):
        try:
            total_duration = stats.stop_time - stats.start_time
            total_duration = _format_seconds(total_duration.total_seconds())
            return total_duration
        except TypeError:
            return ""


def _format_seconds(seconds: float) -> str:
    """Format seconds into a string.

    Args:
        seconds: Seconds to format.

    Returns:
        Formatted string.
    """

    seconds = int(seconds)
    result = [f"{seconds % 60}s"]

    minutes = seconds // 60
    if minutes > 0:
        result.append(f"{minutes % 60}m")
        hours = minutes // 60
        if hours > 0:
            result.append(f"{hours % 24}h")
            days = hours // 24
            if days > 0:
                result.append(f"{days}d")

    return ":".join(reversed(result))


def get_update_date() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc).replace(microsecond=0)



from typing import Protocol

from .async_session import AsyncExperimentSession
from .experiment_session import ExperimentSession


class ExperimentSessionMaker(Protocol):
    """Used to create a new experiment session with predefined parameters."""

    def __call__(self) -> ExperimentSession:
        """Create a new experiment session."""
        ...

    def async_session(self) -> AsyncExperimentSession: ...



import contextlib
import functools
from datetime import datetime
from typing import Callable, Concatenate, TypeVar, ParamSpec, Mapping, Optional, Self

import anyio.to_thread
import attrs
from returns.result import Result
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session

from caqtus.types.iteration import IterationConfiguration
from caqtus.types.parameter import Parameter, ParameterNamespace
from caqtus.types.variable_name import DottedVariableName
from ._experiment_session import _get_global_parameters, _set_global_parameters
from ._path_hierarchy import _does_path_exists, _get_children, _get_path_creation_date
from ._sequence_collection import (
    _get_stats,
    _is_sequence,
    _get_sequence_global_parameters,
    _get_time_lanes,
    _get_iteration_configuration,
    _get_shots,
    _get_shot_parameters,
    _get_shot_end_time,
    _get_shot_start_time,
    _get_shot_data_by_label,
    _get_all_shot_data,
)
from ._serializer import SerializerProtocol
from .. import PureSequencePath
from ..async_session import (
    AsyncExperimentSession,
    AsyncPathHierarchy,
    AsyncSequenceCollection,
)
from ..experiment_session import ExperimentSessionNotActiveError
from ..path_hierarchy import PathNotFoundError, PathIsRootError
from ..sequence_collection import (
    PathIsSequenceError,
    SequenceStats,
    PathIsNotSequenceError,
    PureShot,
)
from ..shot import TimeLanes
from ...types.data import DataLabel, Data

_T = TypeVar("_T")
_P = ParamSpec("_P")


class AsyncSQLExperimentSession(AsyncExperimentSession):
    def __init__(
        self,
        async_session_context: contextlib.AbstractAsyncContextManager[AsyncSession],
        serializer: SerializerProtocol,
    ):
        self._async_session_context = async_session_context
        self._async_session: Optional[AsyncSession] = None
        self.paths = AsyncSQLPathHierarchy(parent_session=self)
        self.sequences = AsyncSQLSequenceCollection(
            parent_session=self, serializer=serializer
        )

    async def __aenter__(self) -> Self:
        if self._async_session is not None:
            error = RuntimeError("Session has already been activated")
            error.add_note(
                "You cannot reactivate a session, you must create a new one."
            )
        self._async_session = await self._async_session_context.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self._async_session_context.__aexit__(exc_type, exc_val, exc_tb)

    async def get_global_parameters(self) -> ParameterNamespace:
        return await self._run_sync(_get_global_parameters)

    async def set_global_parameters(self, parameters: ParameterNamespace) -> None:
        return await self._run_sync(_set_global_parameters, parameters)

    async def _run_sync(
        self,
        fun: Callable[Concatenate[Session, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs
    ) -> _T:
        return await self._session().run_sync(fun, *args, **kwargs)

    def _session(self) -> AsyncSession:
        if self._async_session is None:
            raise ExperimentSessionNotActiveError(
                "Experiment session was not activated"
            )
        return self._async_session


class ThreadedAsyncSQLExperimentSession(AsyncSQLExperimentSession):
    def __init__(
        self,
        session_context: contextlib.AbstractContextManager[Session],
        serializer: SerializerProtocol,
    ):
        self._session_context = session_context
        self._session: Optional[Session] = None

        self.paths = AsyncSQLPathHierarchy(parent_session=self)
        self.sequences = AsyncSQLSequenceCollection(
            parent_session=self, serializer=serializer
        )

    async def __aenter__(self):
        if self._session is not None:
            raise RuntimeError("Session is already active")
        self._session = await anyio.to_thread.run_sync(self._session_context.__enter__)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await anyio.to_thread.run_sync(
            self._session_context.__exit__, exc_type, exc_val, exc_tb
        )

    async def _run_sync(
        self,
        fun: Callable[Concatenate[Session, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs
    ) -> _T:
        if self._session is None:
            raise ExperimentSessionNotActiveError(
                "Experiment session was not activated"
            )
        wrapped = functools.partial(fun, self._session, *args, **kwargs)
        return await anyio.to_thread.run_sync(wrapped)


@attrs.frozen
class AsyncSQLPathHierarchy(AsyncPathHierarchy):
    parent_session: AsyncSQLExperimentSession

    async def does_path_exists(self, path: PureSequencePath) -> bool:
        return await self._run_sync(_does_path_exists, path)

    async def get_children(
        self, path: PureSequencePath
    ) -> Result[set[PureSequencePath], PathNotFoundError | PathIsSequenceError]:
        return await self._run_sync(_get_children, path)

    async def get_path_creation_date(
        self, path: PureSequencePath
    ) -> Result[datetime, PathNotFoundError | PathIsRootError]:
        return await self._run_sync(_get_path_creation_date, path)

    async def _run_sync(
        self,
        fun: Callable[Concatenate[Session, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs
    ) -> _T:
        return await self.parent_session._run_sync(fun, *args, **kwargs)


@attrs.frozen
class AsyncSQLSequenceCollection(AsyncSequenceCollection):
    parent_session: AsyncSQLExperimentSession
    serializer: SerializerProtocol

    async def is_sequence(
        self, path: PureSequencePath
    ) -> Result[bool, PathNotFoundError]:
        return await self._run_sync(_is_sequence, path)

    async def get_stats(
        self, path: PureSequencePath
    ) -> Result[SequenceStats, PathNotFoundError | PathIsNotSequenceError]:
        return await self._run_sync(_get_stats, path)

    async def get_time_lanes(self, path: PureSequencePath) -> TimeLanes:
        return await self._run_sync(_get_time_lanes, path, self.serializer)

    async def get_global_parameters(self, path: PureSequencePath) -> ParameterNamespace:
        return await self._run_sync(_get_sequence_global_parameters, path)

    async def get_iteration_configuration(
        self, path: PureSequencePath
    ) -> IterationConfiguration:
        return await self._run_sync(_get_iteration_configuration, path, self.serializer)

    async def get_shots(
        self, path: PureSequencePath
    ) -> Result[list[PureShot], PathNotFoundError | PathIsNotSequenceError]:
        return await self._run_sync(_get_shots, path)

    async def get_shot_parameters(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DottedVariableName, Parameter]:
        return await self._run_sync(_get_shot_parameters, path, shot_index)

    async def get_all_shot_data(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DataLabel, Data]:
        return await self._run_sync(_get_all_shot_data, path, shot_index)

    async def get_shot_data_by_label(
        self, path: PureSequencePath, shot_index: int, data_label: DataLabel
    ) -> Data:
        return await self._run_sync(
            _get_shot_data_by_label, path, shot_index, data_label
        )

    async def get_shot_start_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime:
        return await self._run_sync(_get_shot_start_time, path, shot_index)

    async def get_shot_end_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime:
        return await self._run_sync(_get_shot_end_time, path, shot_index)

    async def _run_sync(
        self,
        fun: Callable[Concatenate[Session, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs
    ) -> _T:
        return await self.parent_session._run_sync(fun, *args, **kwargs)



from typing import TypeAlias, Protocol, TypeVar

import polars

from caqtus.session import Shot

ShotData: TypeAlias = polars.DataFrame

T = TypeVar("T")


class ShotImporter(Protocol[T]):
    """Protocol for object that can import a value from a shot.

    A shot importer is a callable that takes a shot returns a value of generic type T.
    """

    def __call__(self, shot: Shot) -> T:
        raise NotImplementedError()


DataImporter: TypeAlias = ShotImporter[polars.DataFrame]
LazyDataImporter: TypeAlias = ShotImporter[polars.LazyFrame]



from anyio.abc import ByteSendStream
from anyio.streams.buffered import BufferedByteReceiveStream


async def receive_with_size_prefix(stream: BufferedByteReceiveStream) -> bytes:
    size = await stream.receive_exactly(8)
    size = int.from_bytes(size, "big")
    return await stream.receive_exactly(size)


async def send_with_size_prefix(stream: ByteSendStream, data: bytes) -> None:
    size = len(data).to_bytes(8, "big")
    await stream.send(size + data)



from ._batched import batched
from ._first import first
from ._iterate_in_order import iterate_in_order
from ._pairwise import pairwise

__all__ = ["batched", "first", "pairwise", "iterate_in_order"]



from PySide6.QtCore import QSize
from PySide6.QtWidgets import QLineEdit, QStyleOptionFrame, QStyle


class AutoResizeLineEdit(QLineEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.textChanged.connect(self.resize_to_content)

    def resize_to_content(self):
        text = self.get_text()
        self.resize_to_text(text)

    def get_text(self):
        completer = self.completer()
        if completer is not None:
            if completer.popup().isVisible():
                model = completer.completionModel()
                if model.rowCount() > 0:
                    longest_suggestion = max(
                        (
                            model.data(model.index(index, 0))
                            for index in range(model.rowCount())
                        ),
                        key=lambda x: len(x),
                    )
                    return longest_suggestion
        text = self.text()
        if text:
            return text
        else:
            return self.placeholderText()

    def setCompleter(self, completer):
        super().setCompleter(completer)
        if completer is not None:
            completer.highlighted.connect(self.resize_to_text)

    def resize_to_text(self, text):
        text_size = self.fontMetrics().size(0, text)
        tm = self.textMargins()
        tm_size = QSize(tm.left() + tm.right(), tm.top() + tm.bottom())
        cm = self.contentsMargins()
        cm_size = QSize(cm.left() + cm.right(), cm.top() + cm.bottom())
        extra_size = QSize(8, 4)
        contents_size = text_size + tm_size + cm_size + extra_size
        op = QStyleOptionFrame()
        op.initFrom(self)
        perfect_size = self.style().sizeFromContents(
            QStyle.ContentsType.CT_LineEdit, op, contents_size
        )
        self.setFixedSize(perfect_size)



#!/usr/bin/python
from ..errors import NodeRegistrationError


class NodeFactory(object):
    """
    Node factory that stores all the node types.
    """

    def __init__(self):
        self.__aliases = {}
        self.__names = {}
        self.__nodes = {}

    @property
    def names(self):
        """
        Return all currently registered node type identifiers.

        Returns:
            dict: key=<node name, value=node_type
        """
        return self.__names

    @property
    def aliases(self):
        """
        Return aliases assigned to the node types.

        Returns:
            dict: key=alias, value=node type
        """
        return self.__aliases

    @property
    def nodes(self):
        """
        Return all registered nodes.

        Returns:
            dict: key=node identifier, value=node class
        """
        return self.__nodes

    def create_node_instance(self, node_type=None):
        """
        create node object by the node type identifier or alias.

        Args:
            node_type (str): node type or optional alias name.

        Returns:
            NodeGraphQt.NodeObject: new node object.
        """
        if node_type in self.aliases:
            node_type = self.aliases[node_type]

        _NodeClass = self.__nodes.get(node_type)
        if _NodeClass:
            return _NodeClass()

    def register_node(self, node, alias=None):
        """
        register the node.

        Args:
            node (NodeGraphQt.NodeObject): node object.
            alias (str): custom alias for the node identifier (optional).
        """
        if node is None:
            return

        name = node.NODE_NAME
        node_type = node.type_

        if self.__nodes.get(node_type):
            raise NodeRegistrationError(
                'node type "{}" already registered to "{}"! '
                "Please specify a new plugin class name or __identifier__.".format(
                    node_type, self.__nodes[node_type]
                )
            )
        self.__nodes[node_type] = node

        if self.__names.get(name):
            self.__names[name].append(node_type)
        else:
            self.__names[name] = [node_type]

        if alias:
            if self.__aliases.get(alias):
                raise NodeRegistrationError(
                    'Alias: "{}" already registered to "{}"'.format(
                        alias, self.__aliases.get(alias)
                    )
                )
            self.__aliases[alias] = node_type

    def clear_registered_nodes(self):
        """
        clear out registered nodes, to prevent conflicts on reset.
        """
        self.__nodes.clear()
        self.__names.clear()
        self.__aliases.clear()



"""This module contains utility functions for logging."""

from ._log_call import log_call
from ._log_context import log_cm_decorator, log_async_cm_decorator, log_async_cm
from ._logger import caqtus_logger

__all__ = [
    "caqtus_logger",
    "log_call",
    "log_async_cm",
    "log_cm_decorator",
    "log_async_cm_decorator",
]



import numpy as np

from ._instructions import SequencerInstruction


def with_name(
    instruction: SequencerInstruction, name: str
) -> SequencerInstruction[np.void]:
    """
    Change the dtype of the instruction into a structured array with a single field
    with the given name and the same dtype as the original instruction.
    """

    new_type = np.dtype([(name, instruction.dtype)])
    return instruction.as_type(new_type)
